# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from concurrent.futures import Future
from functools import partial
import json
import requests
from threading import RLock

from bundleplacer.async import submit
from bundleplacer.relationtype import RelationType


class MetadataController:

    def __init__(self, placement_controller, config, error_cb=None):
        self.placement_controller = placement_controller
        self.config = config
        self.error_cb = error_cb
        self.series = placement_controller.bundle.series
        self.charm_names = placement_controller.charm_names()
        # charm_name : charm_metadata full dict
        self.charm_info = {}
        # charm_name : requires/provides lists
        self.iface_info = {}
        self.metadata_future = None
        self.metadata_future_lock = RLock()
        self.charms_providing_iface = defaultdict(list)
        self.charms_requiring_iface = defaultdict(list)
        self.get_recommended_charm_names()
        self.load(self.charm_names + self.recommended_charm_names)

    def get_recommended_charm_names(self):
        if not self.config.getopt('config_filename'):
            self.recommended_charm_names = []
            return
        with open(self.config.getopt('config_filename')) as cf:
            bundle_config = json.load(cf)
        key = self.config.getopt('bundle_key')
        bundles = bundle_config['bundles']

        if isinstance(bundles, list):
            bundle_dict = next((d for d in bundles if d['key'] == key),
                               {})
        else:
            bundle_dict = bundles[key]
        self.recommended_charm_names = bundle_dict.get('recommendedCharms',
                                                       [])

    def load(self, charm_names):
        if self.metadata_future:
            # just wait for successive loads:
            self.metadata_future.result()

        with self.metadata_future_lock:
            self.metadata_future = submit(partial(self._do_load,
                                                  charm_names),
                                          self.handle_search_error)

    def _do_load(self, charm_names):
        ids = "&".join(["id={}/{}".format(self.series, n)
                        for n in charm_names])
        url = 'https://api.jujucharms.com/v4/meta/any?include=charm-metadata&'
        url += ids
        r = requests.get(url)
        metas = r.json()
        for charm_name, charm_dict in metas.items():
            md = charm_dict["Meta"]["charm-metadata"]
            charm_name_no_series = charm_name.split("/")[-1]
            self.charm_info[charm_name_no_series] = charm_dict
            rd = md.get("Requires", {})
            pd = md.get("Provides", {})
            requires = []
            provides = []
            for relname, d in rd.items():
                iface = d["Interface"]
                requires.append((relname, iface))
                self.charms_requiring_iface[iface].append((relname,
                                                           charm_name_no_series))

            for relname, d in pd.items():
                iface = d["Interface"]
                provides.append((relname, iface))
                self.charms_providing_iface[iface].append((relname,
                                                           charm_name_no_series))

            self.iface_info[charm_name_no_series] = dict(requires=requires,
                                                         provides=provides)

    def loaded(self):
        return self.metadata_future.done()

    def add_charm(self, charm_name):
        if charm_name not in self.charm_info:
            self.load([charm_name])

    def get_provides(self, charm_name):
        if not self.loaded():
            return []
        return self.iface_info[charm_name]['provides']

    def get_requires(self, charm_name):
        if not self.loaded():
            return []
        return self.iface_info[charm_name]['requires']

    def get_charm_info(self, charm_name):
        if not self.loaded():
            return None
        return self.charm_info[charm_name]

    def get_services_for_iface(self, iface, reltype):
        services = []
        if reltype == RelationType.Requires:
            cs = self.charms_requiring_iface[iface]
        else:
            cs = self.charms_providing_iface[iface]

        for relname, charm_name in cs:
            pc = self.placement_controller
            services += [(relname, s) for s in
                         pc.services_with_charm(charm_name)]

        return services

    def handle_search_error(self, e):
        self.error_cb(e)


class CharmStoreAPI:
    """Concurrently lookup data from the juju charm store.

    use the get_* functions to get a Future whose result will be the
    requested charm info

    """
    _cache = {}
    _cachelock = RLock()

    def __init__(self, series):
        self.baseurl = 'https://api.jujucharms.com/v4'
        self.series = series

    def _do_remote_lookup(self, charm_name, metakey):
        url = (self.baseurl + '/meta/' +
               'any?include=charm-metadata&id={}/{}'.format(self.series,
                                                            charm_name))
        r = requests.get(url)
        rj = r.json()
        if len(rj.items()) != 1:
            raise Exception("Got wrong number of results from charm store")
        entity = list(rj.values())[0]
        with CharmStoreAPI._cachelock:
            CharmStoreAPI._cache[charm_name] = entity
        return entity

    def _wait_for_pending_lookup(self, f, charm_name, metakey):
        try:
            entity = f.result()
        except:
            return None

        if metakey is None:
            return entity
        return entity['Meta']['charm-metadata'][metakey]

    def _lookup(self, charm_name, metakey, exc_cb):
        with CharmStoreAPI._cachelock:
            if charm_name in CharmStoreAPI._cache:
                val = CharmStoreAPI._cache[charm_name]
                if isinstance(val, Future):
                    f = submit(partial(self._wait_for_pending_lookup,
                                       val, charm_name, metakey),
                               exc_cb)
                else:
                    if metakey is None:
                        d = val
                    else:
                        d = val['Meta']['charm-metadata'][metakey]
                    f = submit(lambda: d, exc_cb)
            else:
                whole_entity_f = submit(partial(self._do_remote_lookup,
                                                charm_name,
                                                metakey),
                                        exc_cb)
                CharmStoreAPI._cache[charm_name] = whole_entity_f

                f = submit(partial(self._wait_for_pending_lookup,
                                   whole_entity_f, charm_name,
                                   metakey),
                           exc_cb)

        return f

    def get_summary(self, charm_name, exc_cb):
        return self._lookup(charm_name, 'Summary', exc_cb)

    def get_entity(self, charm_name, exc_cb):
        return self._lookup(charm_name, None, exc_cb)

    def get_matches(self, substring, exc_cb):
        def _do_search():
            url = (self.baseurl +
                   "/search?text={}&autocomplete=1".format(substring) +
                   "&limit=20&include=charm-metadata&include=bundle-metadata")
            charm_url = url + "&type=charm&series={}".format(self.series)
            bundle_url = url + "&type=bundle"
            cr = requests.get(charm_url)
            crj = cr.json()
            br = requests.get(bundle_url)
            brj = br.json()
            return brj['Results'], crj['Results']

        f = submit(_do_search, exc_cb)
        return f



""" Interfaces to Juju Controllers """

from conjure.juju import Juju


def cloud_type(controller):
    """ Gets the cloud type for controller

    Arguments:
    controller: Specify controller

    Returns:
    String of the cloud-type
    """
    info = Juju.controller_info(controller)
    bootstrap_config = info.get('bootstrap-config', {})
    cloud_type = bootstrap_config.get('cloud-type', None)
    return cloud_type



# -*- mode: python; -*-
#
# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from urwid import WidgetWrap
from bundleplacer.ui import PlacementView
from ubuntui.ev import EventLoop

from bundleplacer.assignmenttype import AssignmentType

log = logging.getLogger('bundleplacer')


class PlacerView(WidgetWrap):
    def __init__(self, placement_controller, config, cb,
                 has_maas=False):
        self.placement_controller = placement_controller
        self.config = config
        self.cb = cb
        self.has_maas = has_maas

        self.selected_machine = None
        self.selected_service = None
        self.pv = PlacementView(
            display_controller=self,
            placement_controller=self.placement_controller,
            config=self.config,
            do_deploy_cb=self.do_deploy,
            has_maas=has_maas)
        super().__init__(self.pv)
        self.pv.reset_selections(top=True)

    def update(self, *args, **kwargs):
        self.pv.update()
        EventLoop.set_alarm_in(1, self.update)

    def status_error_message(self, message):
        pass

    def status_info_message(self, message):
        pass

    def do_deploy(self):
        self.cb()

    def _do_select(self, machine, atype):
        self.placement_controller.assign(machine,
                                         self.selected_service,
                                         atype)
        self.pv.reset_selections()

    def do_select_baremetal(self, machine):
        self._do_select(machine, AssignmentType.BareMetal)

    def do_select_lxd(self, machine):
        self._do_select(machine, AssignmentType.LXD)

    def do_select_kvm(self, machine):
        self._do_select(machine, AssignmentType.KVM)

    def set_selected_service(self, service):
        self.selected_service = service

    def remove_service(self, service):
        self.placement_controller.remove_service(service.service_name)
        self.pv.update()

    def show_default_view(self):
        self.pv.show_default_view()

    def edit_placement(self):
        self.pv.edit_placement()

    def edit_relations(self):
        assert self.selected_service is not None
        self.pv.edit_relations(self.selected_service)

    def clear_selections(self):
        self.pv.clear_selections()



""" Charm utilities

Api for the charmstore:
https://github.com/juju/charmstore/blob/v4/docs/API.md
"""
import yaml
import requests
import os.path as path
from tempfile import NamedTemporaryFile
from conjure.utils import spew

cs = 'https://api.jujucharms.com/v4'


class CharmStoreException(Exception):
    """ CharmStore exception """


def get_bundle(bundle, to_file=False):
    """ Attempts to grab the bundle.yaml

    Arguments:
    bundle: name of bundle
    to_file: store to a temporary file

    Returns:
    Dictionary of bundle's yaml unless to_file is True,
    then returns the path to the downloaded bundle
    """
    bundle = path.join(cs, bundle, 'archive/bundle.yaml')
    req = requests.get(bundle)
    if not req.ok:
        raise CharmStoreException("Problem getting bundle: {}".format(req))
    if to_file:
        with NamedTemporaryFile(mode="w", encoding="utf-8",
                                delete=False) as tempf:
            spew(tempf.name, req.text)
            return tempf.name
    else:
        return yaml.safe_load(req.text)



# Copyright 2015-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging


log = logging.getLogger('bundleplacer')


class Service:
    def __init__(self, service_name, charm_source, summary_future,
                 constraints, depends, conflicts,
                 allowed_assignment_types, num_units, options,
                 allow_multi_units, subordinate, required, relations):
        self.service_name = service_name
        self.charm_name = '-'.join(charm_source.split('/')[-1].split('-')[:-1])
        self.charm_source = charm_source
        self.summary_future = summary_future
        self._summary = "Loading summary…"
        self.constraints = constraints
        self.depends = depends
        self.conflicts = conflicts
        self.allowed_assignment_types = allowed_assignment_types
        self.num_units = num_units
        self.options = options
        self.allow_multi_units = allow_multi_units
        self.subordinate = subordinate
        self.is_core = required
        self.isolate = True if not subordinate else False
        self.relations = relations

    @property
    def summary(self):
        if self.summary_future and self.summary_future.done():
            self._summary = self.summary_future.result()
        return self._summary

    @property
    def display_name(self):
        return "{} ({})".format(self.service_name,
                                self.charm_source)

    def required_num_units(self):
        return self.num_units

    def __repr__(self):
        return "<Service {}>".format(self.service_name)

    def __eq__(self, other):
        me = self.charm_source + self.service_name
        them = other.charm_source + other.service_name
        return me == them

    def __hash__(self):
        """We assume that we won't instantiate multiple Charm objects for the
        same class that have different properties.
        """
        return hash(self.charm_source + self.service_name)



from urwid import ListBox, SimpleListWalker


class SimpleList(ListBox):
    def __init__(self, contents, is_selectable=True):
        self.contents = contents
        self.is_selectable = is_selectable
        self.lbox = self._build_widget()
        super().__init__(self.lbox)

    def _build_widget(self):
        lw = SimpleListWalker([x for x in self.contents])

        return lw

    def selectable(self):
        return self.is_selectable



import logging
from logging.handlers import SysLogHandler


def setup_logging(app, debug=False):
    cmdslog = SysLogHandler('/dev/log',
                            facility=SysLogHandler.LOG_DAEMON)
    if debug:
        env = logging.DEBUG
    else:
        env = logging.INFO
    cmdslog.setLevel(env)
    cmdslog.setFormatter(logging.Formatter(
        "%(name)s: [%(levelname)s] %(message)s"))

    logger = logging.getLogger(app)
    logger.setLevel(env)
    logger.addHandler(cmdslog)
    return logger



""" Application entrypoint
"""

from ubuntui.ev import EventLoop
from ubuntui.palette import STYLES
from conjure.ui import ConjureUI
from conjure.juju import Juju
from conjure import async
from conjure import __version__ as VERSION
from conjure.controllers.welcome import WelcomeController
from conjure.controllers.finish import FinishController
from conjure.controllers.deploysummary import DeploySummaryController
from conjure.controllers.deploy import DeployController
from conjure.controllers.cloud import CloudController
from conjure.controllers.newcloud import NewCloudController
from conjure.controllers.jujucontroller import JujuControllerController
from conjure.controllers.bootstrapwait import BootstrapWaitController
from conjure.controllers.lxdsetup import LXDSetupController
from conjure.log import setup_logging
import json
import sys
import argparse
import os
import os.path as path
import uuid


class ApplicationException(Exception):
    """ Error in application
    """


class ApplicationConfig:
    """ Application config encapsulating common attributes
    used throughout the lifetime of the application.
    """
    def __init__(self):
        # Reference to entire UI
        self.ui = None
        # Global config attr
        self.config = None
        # CLI arguments
        self.argv = None
        # List of all known controllers to be rendered
        self.controllers = None
        # Current Juju model
        self.current_model = None
        # Current controller
        self.current_controller = None
        # Global session id
        self.session_id = os.getenv('CONJURE_TEST_SESSION_ID',
                                    str(uuid.uuid4()))
        # Logger
        self.log = None
        # Environment to pass to processing tasks
        self.env = os.environ.copy()


class Application:
    def __init__(self, argv):
        """ init

        Arguments:
        argv: Options passed in from cli
        """
        self.app = ApplicationConfig()
        self.app.argv = argv
        self.app.ui = ConjureUI()

        with open(argv.build_conf) as json_f:
            config = json.load(json_f)

        with open(argv.build_metadata) as json_f:
            config['metadata_filename'] = path.abspath(argv.build_metadata)
            config['metadata'] = json.load(json_f)

        self.app.config = config

        self.app.controllers = {
            'welcome': WelcomeController(self.app),
            'clouds': CloudController(self.app),
            'newcloud': NewCloudController(self.app),
            'lxdsetup': LXDSetupController(self.app),
            'bootstrapwait': BootstrapWaitController(self.app),
            'deploy': DeployController(self.app),
            'deploysummary': DeploySummaryController(self.app),
            'jujucontroller': JujuControllerController(self.app),
            'finish': FinishController(self.app)
        }

        self.app.log = setup_logging(self.app.config['name'],
                                     self.app.argv.debug)

    def unhandled_input(self, key):
        if key in ['q', 'Q']:
            async.shutdown()
            EventLoop.exit(0)

    def _start(self, *args, **kwargs):
        """ Initially load the welcome screen
        """
        if self.app.argv.status_only:
            self.app.controllers['finish'].render(bundle=None)
        else:
            self.app.controllers['welcome'].render()

    def start(self):
        EventLoop.build_loop(self.app.ui, STYLES,
                             unhandled_input=self.unhandled_input)
        EventLoop.set_alarm_in(0.05, self._start)
        EventLoop.run()


def parse_options(argv):
    parser = argparse.ArgumentParser(prog="conjure-setup")
    parser.add_argument('-c', '--config', dest='build_conf', metavar='CONFIG',
                        help='Path to Conjure config')
    parser.add_argument('-m', '--metadata', dest='build_metadata',
                        metavar='METADATA',
                        help='Path to bundle services metadata')
    parser.add_argument('-s', '--status', action='store_true',
                        dest='status_only',
                        help='Only display the Status of '
                        'existing deployed bundled.')
    parser.add_argument('-d', '--debug', action='store_true',
                        dest='debug',
                        help='Enable debug logging.')
    parser.add_argument(
        '--version', action='version', version='%(prog)s {}'.format(VERSION))
    return parser.parse_args(argv)


def main():
    opts = parse_options(sys.argv[1:])

    if os.geteuid() == 0:
        print("")
        print("This should _not_ be run as root or with sudo.")
        print("")
        sys.exit(1)

    try:
        docs_url = "https://jujucharms.com/docs/stable/getting-started"
        juju_version = Juju.version()
        if int(juju_version[0]) < 2:
            print(
                "Only Juju v2 and above is supported, "
                "your currently installed version is {}.\n\n"
                "Please refer to {} for help on installing "
                "the correct Juju.".format(juju_version, docs_url))
            sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)

    if not opts.build_conf:
        print(
            "A conjure config is required, see conjure-setup -h.")
        sys.exit(1)

    if not path.exists(opts.build_conf):
        print("Unable to find {}".format(opts.build_conf))
        sys.exit(1)

    if not path.exists(opts.build_metadata):
        print("Unable to find {} metadata".format(
            opts.build_metadata
        ))
        sys.exit(1)

    app = Application(opts)
    app.start()



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import os
import sys
import urwid

from bundleplacer import async
from bundleplacer.maas import connect_to_maas

from bundleplacer.config import Config
from bundleplacer.controller import BundleWriter, PlacementController
from bundleplacer.log import setup_logger
from bundleplacer.placerview import PlacerView
from bundleplacer.fixtures.maas import FakeMaasState
from ubuntui.ev import EventLoop
from ubuntui.palette import STYLES
from ubuntui.anchors import Header, Footer

log = None


class PlacerUI(urwid.Frame):
    def __init__(self, placerview):
        super().__init__(body=placerview, header=Header(), footer=Footer())


def parse_options(argv):
    parser = argparse.ArgumentParser(description='Juju Bundle Placer',
                                     prog='placer',
                                     argument_default=argparse.SUPPRESS)
    parser.add_argument("bundle_filename", metavar='bundle',
                        help="Bundle file to edit")
    parser.add_argument("--metadata", dest="metadata_filename",
                        metavar='metadatafile',
                        help="Optional metadata file describing constraints "
                        "on services in bundle")
    parser.add_argument("--fake-maas", dest="fake_maas",
                        action="store_true", default=False)
    parser.add_argument("--maas-ip", dest="maas_ip", default=None)
    parser.add_argument("--maas-cred", dest="maas_cred", default=None)
    parser.add_argument("-o", dest="out_filename", default=None)
    return parser.parse_args(argv)


def main():
    opts = parse_options(sys.argv[1:])

    config = Config('bundle-placer', opts.__dict__)
    config.save()

    setup_logger(cfg_path=config.cfg_path)
    log = logging.getLogger('bundleplacer')
    log.debug(opts.__dict__)

    log.info("Editing file: {}".format(opts.bundle_filename))

    if opts.maas_ip and opts.maas_cred:
        creds = dict(api_host=opts.maas_ip,
                     api_key=opts.maas_cred)
        maas, maas_state = connect_to_maas(creds)
    elif opts.fake_maas:
            maas = None
            maas_state = FakeMaasState()
    else:
        maas = None
        maas_state = None

    placement_controller = PlacementController(config=config,
                                               maas_state=maas_state)

    def cb():
        if maas:
            maas.tag_name(maas.nodes)

        bw = BundleWriter(placement_controller)
        if opts.out_filename:
            outfn = opts.out_filename
        else:
            path, ext = os.path.splitext(opts.bundle_filename)
            outfn = "{}-out{}".format(path, ext)
        bw.write_bundle(outfn)
        async.shutdown()
        raise urwid.ExitMainLoop()

    has_maas = (maas_state is not None)
    mainview = PlacerView(placement_controller, config, cb, has_maas=has_maas)
    ui = PlacerUI(mainview)

    def unhandled_input(key):
        if key in ['q', 'Q']:
            async.shutdown()
            raise urwid.ExitMainLoop()
    EventLoop.build_loop(ui, STYLES, unhandled_input=unhandled_input)
    mainview.loop = EventLoop.loop
    mainview.update()
    EventLoop.run()



from conjure.ui.views.bootstrapwait import BootstrapWaitView
from ubuntui.ev import EventLoop


class BootstrapWaitController:
    def __init__(self, app):
        self.app = app

    def refresh(self, *args):
        self.view.redraw_kitt()
        EventLoop.set_alarm_in(1, self.refresh)

    def render(self):
        self.view = BootstrapWaitView(self.app)
        self.app.ui.set_header(
            title="Initializing model",
            excerpt="Please wait while Juju bootstraps the model.",
        )
        self.app.ui.set_body(self.view)
        self.app.ui.set_subheader("Press (Q) to cancel bootstrap and exit.")
        EventLoop.set_alarm_in(1, self.refresh)



from ubuntui.utils import Color, Padding
from ubuntui.widgets.hr import HR
from ubuntui.widgets.text import Instruction
from ubuntui.widgets.input import StringEditor
from ubuntui.widgets.buttons import (cancel_btn, confirm_btn, PlainButton)
from urwid import (WidgetWrap, Pile,
                   Text, Filler, Columns)
from conjure.models.bundle import BundleModel
from conjure.api.controllers import cloud_type


class JujuControllerView(WidgetWrap):
    controller_mode = None

    def __init__(self, app, models=None, cb=None):
        self.app = app
        self.cb = cb
        self.models = models
        self.input_new_controller = StringEditor()
        self.group = []
        self.config = self.app.config

        if self.models is not None:
            self.controller_mode = 'existing'
            super().__init__(self._build_existingcontroller_widget())
        else:
            self.controller_mode = 'new'
            super().__init__(self._build_newcontroller_widget())

    def _swap_focus(self):
        w_count = len(self._w.body.contents) - 1
        pos = self._w.body.focus_position
        if self.controller_mode == 'existing':
            if pos >= 3 and pos < w_count:
                self._w.body.focus_position = w_count
            else:
                self._w.body.focus_position = 3
        else:
            if pos >= 2 and pos < w_count:
                self._w.body.focus_position = w_count
            else:
                self._w.body.focus_position = 2

    def keypress(self, size, key):
        if key in ['tab', 'shift tab']:
            self._swap_focus()
        return super().keypress(size, key)

    def _build_buttons(self):
        cancel = cancel_btn(on_press=self.cancel)
        buttons = [
            Padding.line_break(""),
            Color.button_secondary(cancel,
                                   focus_map='button_secondary focus'),
        ]
        return Pile(buttons)

    def _build_newcontroller_widget(self):
        items = [
            Padding.center_60(Instruction("Please fill out the input below:")),
            Padding.center_60(HR()),
            Padding.center_60(
                Columns(
                    [
                        ('weight', 0.2, Text("Model")),
                        Color.string_input(self.input_new_controller,
                                           focus_map='string_input focus'),
                    ]
                )),
            Padding.center_60(HR()),
            Padding.center_20(self._build_buttons())
        ]
        return Filler(Pile(items), valign="middle")

    def _build_existingcontroller_widget(self):
        items = [
            Padding.center_60(
                Instruction(
                    "Please select a model below or optionally "
                    "create a new model:")),
            Padding.center_60(HR())
        ]
        for k in sorted(self.models.keys()):
            controller_name = k.split('local.').pop()
            controller_type = cloud_type(controller_name)
            items.append(Padding.center_60(
                Instruction("Model: {} ({})".format(controller_name,
                                                    controller_type))))
            for m in self.models[k]['models']:
                if m['name'] == 'admin':
                    continue
                items.append(
                    Padding.center_50(
                        Color.body(
                            confirm_btn(label="{}:{}".format(controller_name,
                                                             m['name']),
                                        on_press=self.submit),
                            focus_map='menu_button focus'
                        )
                    )
                )
            items.append(Padding.line_break(""))

        # If the bundle only supports MAAS dont provide the option to connect
        # to other public clouds.
        if BundleModel.whitelist() and "maas" in BundleModel.whitelist():
            items.append(Padding.line_break(""))
            items.append(Padding.center_50(
                Color.body(
                    PlainButton(label="Connect to an existing MAAS",
                                on_press=self.submit_new_maas),
                    focus_map='menu_button focus')
            ))
        else:
            items.append(Padding.center_50(
                Color.body(
                    PlainButton(label="Select a new Cloud Provider",
                                on_press=self.select_cloud),
                    focus_map='menu_button focus')
            ))

        items.append(
            Padding.center_60(HR()))
        items.append(Padding.center_20(self._build_buttons()))
        return Filler(Pile(items), valign="middle")

    def select_cloud(self, btn):
        return self.app.controllers['clouds'].render()

    def submit_new_maas(self, btn):
        return self.app.controllers['newcloud'].render("maas")

    def submit(self, result):
        self.cb(result.label)

    def cancel(self, btn):
        self.cb(back=True)



""" Deployment summary view


List out the updated bundle in a cleaner view showing what
charms and their relations will be done.
"""

from urwid import WidgetWrap, Text, Pile, Filler
from ubuntui.widgets.buttons import PlainButton
from ubuntui.widgets.text import Instruction
from ubuntui.widgets.hr import HR
from ubuntui.utils import Color, Padding
import yaml


class DeploySummaryView(WidgetWrap):
    def __init__(self, app, bundle, cb):
        """ Init

        Arguments:
        app: app config
        bundle: dictionary of bundle
        cb: callback
        """
        self.app = app
        with open(bundle) as b:
            self.bundle = yaml.load(b)
        self.cb = cb
        _pile = [
            Padding.center_90(
                Instruction(
                    "Services and Relations to be deployed and set.")),
            Padding.center_90(HR()),
            Padding.center_90(self.build_summary()),
            Padding.line_break(""),
            Padding.center_20(self._build_buttons())
        ]
        super().__init__(Filler(Pile(_pile), valign="top"))

    def _build_buttons(self):
        deploy = PlainButton(label="Deploy", on_press=self.done)
        cancel = PlainButton(label="Cancel", on_press=self.cancel)
        buttons = [
            Color.button_primary(deploy,
                                 focus_map='button_primary focus'),
            Color.button_secondary(cancel,
                                   focus_map='button_secondary focus')
        ]
        return Pile(buttons)

    def build_summary(self):
        """ Builds out the summary of the bundle
        """
        rows = []
        rows.append(
            Text("You will be deploying these services: {}.  ".format(
                ", ".join(self.bundle['services'].keys())
            )))

        machines = None
        if not self.bundle['machines']:
            machines = ("They will be autoplaced across multiple containers.")
        else:
            machines = ("They will be placed across {} machine(s).".format(
                len(self.bundle['machines'].keys())
            ))
        rows.append(Padding.line_break(""))
        rows.append(Text(machines))
        rows.append(Padding.line_break(""))

        return Pile(rows)

    def done(self, button):
        self.cb()

    def cancel(self, button):
        self.cb(back=True)



import os
import sys
import yaml
import argparse
from code import interact


def parse_options(argv):
    parser = argparse.ArgumentParser(description='Macumba Shell',
                                     prog='macumba-shell')
    parser.add_argument('--v1', action='store_true', dest='v1',
                        help='Use Juju 1.x API')
    parser.add_argument('--v2', action='store_true', dest='v2',
                        help='Use Juju 2.x API')
    parser.add_argument('-m', '--model', dest='model',
                        help='The Environment(v1)/Model(v2) to connect to.')
    return parser.parse_args(argv)


def main():
    opts = parse_options(sys.argv[1:])
    if not opts.model:
        raise Exception("Must choose a Environment/Model.")
    if opts.v1:
        juju_home = os.getenv("JUJU_HOME", "~/.juju")
        from .v1 import JujuClient  # noqa
        env = os.path.expanduser(
            os.path.join(
                juju_home,
                "environments/{}.jenv".format(opts.model)))
        if not os.path.isfile(env):
            raise Exception("Unable to locate: {}".format(env))
        env_yaml = yaml.load(open(env))
        uuid = env_yaml['environ-uuid']
        server = env_yaml['state-servers'][0]
        password = env_yaml['password']
        user = env_yaml['user']
        url = os.path.join('wss://', server, 'environment', uuid, 'api')

    elif opts.v2:
        xdg_home = os.getenv("XDG_DATA_HOME", "~/.local/share")
        juju_home = os.path.join(xdg_home, 'juju')
        from .v2 import JujuClient  # noqa
        env = os.path.expanduser(
            os.path.join(
                juju_home,
                "models/cache.yaml"))
        if not os.path.isfile(env):
            raise Exception("Unable to locate: {}".format(env))

        env = yaml.load(open(env))
        uuid = env['server-user'][opts.model]['server-uuid']
        server = env['server-data'][uuid]['api-endpoints'][0]
        password = env['server-data'][uuid]['identities']['admin']
        url = os.path.join('wss://', server, 'model', uuid, 'api')
    else:
        raise Exception("Could not determine Juju API Version to use.")

    print('Connecting to {}'.format(url))
    j = JujuClient(url=url, password=password)
    j.login()

    interact(banner="juju client logged in. Object is named 'j',"
             " so j.status() will fetch current status as a dict.",
             local=locals())



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging


from urwid import AttrMap, Columns, Divider, Pile, Text, WidgetWrap

from ubuntui.widgets.buttons import MenuSelectButton, PlainButton

from bundleplacer.assignmenttype import AssignmentType

log = logging.getLogger('bundleplacer')


class SimpleMachineWidget(WidgetWrap):

    """A widget displaying a machine. When selected, shows action buttons
    for placement types.

    machine - the machine to display

    controller - a PlacementController instance

    display_controller - a PlacerView instance

    show_assignments - display info about which charms are assigned
    and what assignment type (LXC, KVM, etc) they have.

    """

    def __init__(self, machine, controller,
                 display_controller, show_assignments=True):
        self.machine = machine
        self.controller = controller
        self.display_controller = display_controller
        self.show_assignments = show_assignments
        self.is_selected = False
        w = self.build_widgets()
        super().__init__(w)
        self.update()

    def selectable(self):
        return True

    def build_widgets(self):

        self.button = MenuSelectButton("I AM A MACHINE", self.do_select)
        self.action_button_cols = Columns([])
        self.action_buttons = []

        self.pile = Pile([self.button])
        return self.pile

    def update_machine(self):
        """Refresh with potentially updated machine info from controller.
        Assumes that machine exists - machines going away is handled
        in machineslist.update().
        """
        self.machine = next((m for m in self.controller.machines()
                             if m.instance_id == self.machine.instance_id),
                            None)

    def update(self):
        self.update_machine()
        self.update_action_buttons()

        if self.is_selected:
            self.update_selected()
        else:
            self.update_unselected()

    def info_markup(self):
        m = self.machine
        return [self.machine.hostname + "\n",
                'arch: {}  '.format(m.arch),
                'cores: {}  '.format(m.cpu_cores),
                'mem: {}  '.format(m.mem),
                'storage: {}'.format(m.storage)]

    def update_selected(self):
        cn = self.display_controller.selected_service.service_name
        msg = Text("  Add {} to {}:".format(cn,
                                            self.machine.hostname))
        self.pile.contents = [(msg, self.pile.options()),
                              (self.action_button_cols,
                               self.pile.options()),
                              (Divider(), self.pile.options())]

    def update_unselected(self):
        markup = self.info_markup()
        self.button.set_label(markup)
        self.pile.contents = [(AttrMap(self.button, 'text',
                                       'button_secondary focus'),
                               self.pile.options()),
                              (Divider(), self.pile.options())]

    def update_action_buttons(self):

        all_actions = [(AssignmentType.BareMetal,
                        'Add as Bare Metal',
                        self.select_baremetal),
                       (AssignmentType.LXD,
                        'Add as LXD',
                        self.select_lxd),
                       (AssignmentType.KVM,
                        'Add as KVM',
                        self.select_kvm)]

        sc = self.display_controller.selected_service
        if sc:
            allowed_set = set(sc.allowed_assignment_types)
            allowed_types = set([atype for atype, _, _ in all_actions])
            allowed_types = allowed_types.intersection(allowed_set)
        else:
            allowed_types = set()

        # + 1 for the cancel button:
        if len(self.action_buttons) == len(allowed_types) + 1:
            return

        self.action_buttons = [AttrMap(PlainButton(label,
                                                   on_press=func),
                                       'button_secondary',
                                       'button_secondary focus')
                               for atype, label, func in all_actions
                               if atype in allowed_types]
        self.action_buttons.append(
            AttrMap(PlainButton("Cancel",
                                on_press=self.do_cancel),
                    'button_secondary',
                    'button_secondary focus'))

        opts = self.action_button_cols.options()
        self.action_button_cols.contents = [(b, opts) for b in
                                            self.action_buttons]

    def do_select(self, sender):
        if self.display_controller.selected_service is None:
            return
        self.is_selected = True
        self.update()
        self.pile.focus_position = 1
        self.action_button_cols.focus_position = 0

    def do_cancel(self, sender):
        self.is_selected = False
        self.update()
        self.pile.focus_position = 0

    def _do_select_assignment(self, atype):
        {AssignmentType.BareMetal:
         self.display_controller.do_select_baremetal,
         AssignmentType.LXD:
         self.display_controller.do_select_lxd,
         AssignmentType.KVM:
         self.display_controller.do_select_kvm}[atype](self.machine)
        self.pile.focus_position = 0

    def select_baremetal(self, sender):
        self._do_select_assignment(AssignmentType.BareMetal)

    def select_lxd(self, sender):
        self._do_select_assignment(AssignmentType.LXD)

    def select_kvm(self, sender):
        self._do_select_assignment(AssignmentType.KVM)



import shutil
import os
from subprocess import check_call, CalledProcessError
from conjure.models.bundle import BundleModel
from conjure.async import submit


class UtilsException(Exception):
    """ Error in utils
    """
    pass


def install_home():
    """ returns installer user home
    """
    return os.path.expanduser("~" + install_user())


def juju_path():
    """ returns juju path for $user
    """
    return os.getenv('JUJU_DATA',
                     os.path.expanduser('~/.local/share/juju'))


def mkdir(path):
    if not os.path.isdir(path):
        os.makedirs(path)
        chown(path, install_user(), recursive=True)


def chown(path, user, group=None, recursive=False):
    """ Change user/group ownership of file

    Arguments:
    path: path of file or directory
    user: new owner username
    group: new owner group name
    recursive: set files/dirs recursively
    """
    if group is None:
        group = user

    try:
        if not recursive or os.path.isfile(path):
            shutil.chown(path, user, group)
        else:
            for root, dirs, files in os.walk(path):
                shutil.chown(root, user, group)
                for item in dirs:
                    shutil.chown(os.path.join(root, item), user, group)
                for item in files:
                    shutil.chown(os.path.join(root, item), user, group)
    except OSError as e:
        raise UtilsException(e)


def spew(path, data, owner=None):
    """ Writes data to path
    Arguments:
    path: path of file to write to
    data: contents to write
    owner: optional owner of file
    """
    with open(path, 'w') as f:
        f.write(data)
    if owner:
        try:
            chown(path, owner)
        except:
            raise UtilsException(
                "Unable to set ownership of {}".format(path))


def slurp(path):
    """ Reads data from path

    Arguments:
    path: path of file
    """
    try:
        with open(path) as f:
            return f.read().strip()
    except IOError:
        raise IOError


def install_user():
    """ returns sudo user
    """
    user = os.getenv('SUDO_USER', None)
    if not user:
        user = os.getenv('USER', 'root')
    return user


def pollinate(session, tag, log):
    """ fetches random seed

    Tag definitions:
        W001 - welcome shown
        B001 - bundle selected
        CS - cloud selected
        CC - cloud creation started
        CA - cloud credentials added
        L001 - LXD Setup started
        L002 - LXD Setup completed
        J001 - juju post-bootstrap started
        J002 - juju post-bootstrap completed
        J003 - juju bootstrap started
        J004 - juju bootstrap completed
        CS - controller selected
        PM - placement/bundle editor shown (maas)
        PS - placement/bundle editor shown (other)
        PC - placements committed
        SS - deploy summary shown
        DS - deploy started
        DC - deploy complete
        XA - pre processing started
        XB - post processing started

        UC - user cancelled
        EC - error getting credentials
        EP - error in placement/bundle editor
        EB - error juju bootstrap
        ED - error deploying
        E001 - error in post bootstrap phase
        E002 - error in post processor
        E003 - error in pre processor

    Arguments:
    session: randomly generated session id
    tag: custom tag
    log: logger
    """
    if not os.path.isfile('/usr/bin/pollinate'):
        log.warning("pollinate binary not found")
        return
    bundle_key = BundleModel.key()
    if not bundle_key:
        bundle_key = '-'
    agent_str = 'conjure/{}/{}/{}'.format(session, bundle_key, tag)

    def do_pollinate():
        try:
            cmd = ("sudo su - -c 'pollinate -q -r --curl-opts "
                   "\"-k --user-agent {}\"'".format(agent_str))
            log.debug("pollinate: {}".format(cmd))
            check_call(cmd, shell=True)
        except CalledProcessError as e:
            log.warning("Generating random seed failed: {}".format(e))
    submit(do_pollinate, lambda _: None)



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import yaml

from bundleplacer.assignmenttype import AssignmentType, label_to_atype
from bundleplacer.service import Service

log = logging.getLogger('bundleplacer')


class keydict(dict):
    def __missing__(self, key):
        return key


def create_service(servicename, service_dict, servicemeta, relations):

    # a little cleaning to normalize a dict from the charmstore v4 api:
    service_dict = {k.lower(): v for k, v in service_dict.items()}
    if 'num_units' not in service_dict:
        service_dict['num_units'] = service_dict.get('numunits', 0)

    # some attempts to guess at subordinate status from bundle format,
    # to avoid having to include it in metadata:

    # This doesn't work because bundles with no machines might
    # just use the juju default:

    # is_subordinate = 'to' not in service_dict.keys()

    is_subordinate = service_dict.get('num_units', 0) == 0

    charm_name = service_dict['charm'].split('/')[-1]
    charm_name = '-'.join(charm_name.split('-')[:-1])

    myrelations = []
    for src, dst in relations:
        if src.startswith(servicename) or dst.startswith(servicename):
            myrelations.append((src, dst))

    service = Service(service_name=servicename,
                      charm_source=service_dict['charm'],
                      summary_future=None,
                      constraints=servicemeta.get('constraints', {}),
                      depends=servicemeta.get('depends', []),
                      conflicts=servicemeta.get('conflicts', []),
                      allowed_assignment_types=servicemeta.get(
                          'allowed_assignment_types',
                          list(AssignmentType)),
                      num_units=service_dict.get('num_units', 1),
                      options=service_dict.get('options', {}),
                      allow_multi_units=servicemeta.get('allow_multi_units',
                                                        True),
                      subordinate=is_subordinate,
                      required=servicemeta.get('required', True),
                      relations=myrelations)

    # Make sure to map any strings to an assignment type enum
    if any(isinstance(atype, str)
           for atype in service.allowed_assignment_types):
        service.allowed_assignment_types = label_to_atype(
            service.allowed_assignment_types)
    return service


class BundleMergeException(Exception):
    """Error merging two bundles"""


class Bundle:
    def __init__(self, filename=None, metadatafilename=None,
                 bundle_data=None, metadata=None):
        assert(filename or bundle_data)
        self.filename = filename
        self.metadatafilename = metadatafilename
        if self.filename:
            with open(self.filename) as f:
                self._bundle = yaml.load(f)
        else:
            self._bundle = bundle_data
        if metadatafilename:
            with open(self.metadatafilename) as f:
                self._metadata = yaml.load(f)
        elif metadata:
            self._metadata = metadata
        else:
            self._metadata = {}

        self._bundle = {k.lower(): v for
                        k, v in self._bundle.items()}
        for k in ['machines', 'services']:
            for name, val in self._bundle.get(k, {}).items():
                self._bundle[k][name] = {key.lower(): v for
                                         key, v in val.items()}

        if 'services' not in self._bundle.keys():
            raise Exception("Invalid Bundle.")

    def add_new_service(self, charm_name, charm_dict, service_name=None):
        if service_name is None:
            i = 1
            service_name = charm_name
            while service_name in self._bundle['services']:
                service_name = "{}-{}".format(charm_name, i)
                i += 1

        new_dict = {'charm': charm_dict['Id'],
                    'num_units': 1}
        self._bundle['services'][service_name] = new_dict

    def remove_service(self, service_name):
        if service_name in self._bundle['services']:
            del self._bundle['services'][service_name]

        for r1, r2 in self._bundle['relations']:
            s1 = r1.split(':')[0]
            s2 = r2.split(':')[0]
            if s1 == service_name or s2 == service_name:
                self._bundle['relations'].remove([r1, r2])

    def add_relation(self, s1_name, s1_rel, s2_name, s2_rel):
        r = ["{}:{}".format(s1_name, s1_rel),
             "{}:{}".format(s2_name, s2_rel)]
        self._bundle['relations'].append(r)

    def remove_relation(self, s1_name, s1_rel, s2_name, s2_rel):
        r = self.find_relation(s1_name, s1_rel, s2_name, s2_rel)
        self._bundle['relations'].remove(r)

    def is_related(self, s1_name, s1_rel, s2_name, s2_rel):
        """Checks if a relation exists. If the relation in the bundle does not
        specify relation names, this returns true for any relation names.
        """
        r = self.find_relation(s1_name, s1_rel, s2_name, s2_rel)
        return r is not None

    def find_relation(self, s1_name, s1_rel, s2_name, s2_rel):
        a = "{}:{}".format(s1_name, s1_rel)
        b = "{}:{}".format(s2_name, s2_rel)
        rels = self._bundle['relations']
        for x in [[a, b], [b, a], [s1_name, s2_name], [s2_name, s1_name]]:
            if x in rels:
                return x
        return None

    @property
    def services(self):
        services = []
        metadata = self._metadata.get('services', {})
        bundle_services = self._bundle.get('services', {})
        relations = self._bundle.get('relations', [])
        for servicename, sd in bundle_services.items():
            sm = metadata.get(servicename, {})
            services.append(create_service(servicename, sd,
                                           sm, relations))
        return services

    @property
    def machines(self):
        return self._bundle.get('machines', {})

    @property
    def series(self):
        return self._bundle.get('series', 'trusty')

    def clear_machines_and_placement(self):
        self._bundle['machines'] = {}
        for sname, sd in self._bundle['services'].items():
            if 'to' in sd:
                del(sd['to'])

    @property
    def assignments(self):
        """returns a dict of service_name: [to-strings]"""
        assert 'services' in self._bundle
        assignments = {}
        for sname, sd in self._bundle['services'].items():
            if 'to' in sd:
                assignments[sname] = sd['to']
        return assignments

    def extra_items(self):
        return {k: v for k, v in self._bundle.items()
                if k not in ['services', 'machines', 'relations']}

    def update(self, other_bundle):
        """Merges one bundle with another, renaming any duplicate service
        names or machine names.

        returns a pair: (new_machines, new_services, new_assignments)
        """
        new_machines = {}
        new_service_names = []
        new_assignments = {}

        assert 'services' in self._bundle

        # check for conflicts that can't be resolved by renaming
        for k, v in self._bundle.items():
            if k in ['services', 'machines', 'relations']:
                continue
            if self._bundle[k] != other_bundle._bundle[k]:
                m = "Can't merge top level key {}".format(k)
                raise BundleMergeException(m)

        service_renames = keydict()
        for sname, sd in other_bundle._bundle['services'].items():
            if sname in self._bundle['services']:
                newname = sname + "-1"
                service_renames[sname] = newname

        # generate machine renames and merge machines
        machine_renames = keydict()
        n_machines_mine = len(self._bundle.get('machines', {}))
        other_machines = other_bundle._bundle.get('machines', {})
        for mname, md in other_machines.items():
            if mname in self._bundle['machines']:
                newname = str(int(mname) + n_machines_mine)
                machine_renames[mname] = newname
            self._bundle['machines'][machine_renames[mname]] = md
            new_machines[machine_renames[mname]] = md

        # apply service renames to relations
        def rename_relation(r, rd):
            parts = r.split(":")
            if len(parts) > 1:
                return "{}:{}".format(rd[parts[0]],
                                      parts[1])
            else:
                return rd[parts[0]]

        for (or1, or2) in other_bundle._bundle['relations']:
            nr1 = rename_relation(or1, service_renames)
            nr2 = rename_relation(or2, service_renames)
            if [nr1, nr2] in self._bundle['relations'] or\
               [nr2, nr1] in self._bundle['relations']:
                continue
            self._bundle['relations'].append([nr1, nr2])

        # apply machine renames to services
        def rename_machine(to, md):
            parts = to.split(":")
            if len(parts) > 1:
                return "{}:{}".format(parts[0],
                                      md[parts[1]])
            else:
                return md[parts[0]]

        for sname, sd in other_bundle._bundle['services'].items():
            new_sd = sd.copy()
            if 'to' in sd:
                new_sd['to'] = [rename_machine(to, machine_renames)
                                for to in sd['to']]
            self._bundle['services'][service_renames[sname]] = new_sd
            new_service_names.append(service_renames[sname])
            if 'to' in sd:
                new_assignments[service_renames[sname]] = new_sd['to']

        new_services = [s for s in self.services
                        if s.service_name in new_service_names]

        return new_machines, new_services, new_assignments



from __future__ import unicode_literals
from urwid import (WidgetWrap, Text, Pile,
                   Columns, Filler)
from ubuntui.widgets.hr import HR
from ubuntui.widgets.buttons import (cancel_btn, menu_btn)
from ubuntui.utils import Color, Padding
from ubuntui.ev import EventLoop
from conjure.utils import pollinate


class WelcomeView(WidgetWrap):
    def __init__(self, app, cb):
        self.app = app
        self.cb = cb
        self.current_focus = 2
        _pile = [
            Padding.center_90(
                Color.info_context(Text("Choose a solution to get started:"))),
            Padding.center_90(HR()),
            Padding.center_90(self.build_menuable_items()),
            Padding.line_break(""),
            Padding.center_20(self.buttons())
        ]
        super().__init__(Filler(Pile(_pile), valign="top"))

    def _swap_focus(self):
        if self._w.body.focus_position == 2:
            self._w.body.focus_position = 4
        else:
            self._w.body.focus_position = 2

    def keypress(self, size, key):
        if key in ['tab', 'shift tab']:
            self._swap_focus()
        return super().keypress(size, key)

    def buttons(self):
        cancel = cancel_btn(on_press=self.cancel)

        buttons = [
            Color.button_secondary(cancel, focus_map='button_secondary focus')
        ]
        return Pile(buttons)

    def build_menuable_items(self):
        """ Builds a list of bundles available to install
        """
        bundles = self.app.config['bundles']
        cols = []
        for bundle in bundles:
            cols.append(
                Columns(
                    [
                        ("weight", 0.2, Color.body(
                            menu_btn(label=bundle['name'],
                                     on_press=self.done),
                            focus_map="menu_button focus")),
                        ("weight", 0.3, Text(bundle['summary'],
                                             align="left"))
                    ],
                    dividechars=1
                )
            )
            cols.append(Padding.line_break(""))
        return Pile(cols)

    def cancel(self, button):
        pollinate(self.app.session_id, 'UC', self.app.log)
        EventLoop.exit(0)

    def done(self, result):
        self.cb(result.label)



""" Juju helpers
"""
from conjure.shell import shell
from conjure.utils import juju_path
import os
import yaml
import json
from macumba.v2 import JujuClient
from macumba.errors import LoginError
from functools import wraps, partial
from conjure import async


class JujuNotFoundException(Exception):
    """ The available config is not found """


class JujuModelUserNotFound(Exception):
    """ Unable to find user associated with model """


class JujuControllerNotFound(Exception):
    """ Unable to find controller """


class JujuModelNotFound(Exception):
    """ Unable to find models """


class JujuCloudNotFound(Exception):
    """ Unable to determine cloud """


class JujuConfigNotFound(Exception):
    """ Unable to load config file """


# login decorator
def requires_login(f):
    def _decorator(*args, **kwargs):
        if not Juju.is_authenticated:
            Juju.login()
        return f(*args, **kwargs)
    return wraps(f)(_decorator)


def read_config(name):
    """ Reads a juju config file

    Arguments:
    name: filename without extension (ext defaults to yaml)

    Returns:
    dictionary of yaml object
    """
    abs_path = os.path.join(juju_path(), "{}.yaml".format(name))
    if not os.path.isfile(abs_path):
        raise JujuConfigNotFound("Cannot load {}".format(abs_path))
    return yaml.safe_load(open(abs_path))


def current_controller():
    c_path = os.path.join(juju_path(), 'current-controller')
    if not os.path.isfile(c_path):
        return None
    with open(c_path) as fp:
        out = fp.read().strip()
        return out


class Juju:
    is_authenticated = False
    client = None
    user_tag = None

    @classmethod
    def login(cls, force=False):
        """ Login to Juju API server
        """
        if not cls.available():
            return

        if cls.is_authenticated is True and not force:
            return

        current_controller = cls.current_controller()
        if not current_controller:
            raise LoginError("Unable to determine current controller")

        env = cls.controller(current_controller)
        account = cls.account(current_controller)
        uuid = cls.model(cls.current_model())['model-uuid']
        server = env['api-endpoints'][0]
        cls.user_tag = "user-{}".format(account['current'])
        current_user = account['current']
        password = account['users'][current_user]['password']
        url = os.path.join('wss://', server, 'model', uuid, 'api')
        cls.client = JujuClient(
            user=cls.user_tag,
            url=url,
            password=password)
        try:
            cls.client.login()
        except LoginError as e:
            raise e
        cls.is_authenticated = True

    @classmethod
    def bootstrap(cls, controller, cloud, series=None, log=None):
        """ Performs juju bootstrap

        If not LXD pass along the newly defined credentials

        Arguments:
        controller: name of your controller
        cloud: name of local or public cloud to deploy to
        series: define the bootstrap series defaults to xenial
        log: application logger
        """
        cmd = "juju bootstrap {} {} --upload-tools ".format(
                  controller, cloud, series)
        if series is not None:
            cmd += "--bootstrap-series={} ".format(series)
        if cloud != "lxd" and cloud != "localhost":
            cmd += "--credential {}".format(controller)
        if log:
            log.debug("bootstrap cmd: {}".format(cmd))
        return shell(cmd)

    @classmethod
    def bootstrap_async(cls, controller, cloud,
                        series=None, log=None, exc_cb=None):
        """ Performs a bootstrap asynchronously
        """
        return async.submit(partial(cls.bootstrap, controller,
                                    cloud, series, log), exc_cb)

    @classmethod
    def log(cls, limit=1):
        """ returns juju debug-log output
        """
        sh = shell('juju debug-log -T')
        return sh

    @classmethod
    def available(cls):
        """ Checks if juju is available

        Returns:
        True/False if juju status was successful and a environment is found
        """
        return 0 == shell('juju status').code

    @classmethod
    def autoload_credentials(cls):
        """ Automatically checks known places for cloud credentials
        """
        return 0 == shell('juju autoload-credentials').code

    @classmethod
    def credential(cls, cloud, user):
        """ Get credentials for user

        Arguments:
        cloud: cloud applicable to user credentials
        user: user listed in the credentials
        """
        creds = cls.credentials()
        if cloud in creds.keys():
            if user in creds[cloud].keys():
                return creds[cloud][user]
        raise JujuModelUserNotFound(
            "Unable to locate credentials for: {}".format(user))

    @classmethod
    def credentials(cls, secrets=True):
        """ List credentials

        This will fallback to reading the credentials file directly

        Arguments:
        secrets: True/False whether to show secrets (ie password)

        Returns:
        List of credentials
        """
        cmd = 'juju list-credentials --format yaml'
        if secrets:
            cmd += ' --show-secrets'
        sh = shell(cmd)
        if sh.code > 0:
            try:
                env = read_config('credentials')
                return env['credentials']
            except:
                raise JujuNotFoundException(
                    "Unable to list credentials: {}".format(sh.errors()))
        env = yaml.safe_load("\n".join(sh.output()))
        return env['credentials']

    @classmethod
    def clouds(cls):
        """ List available clouds

        Returns:
        Dictionary of all known clouds including newly created MAAS/Local
        """
        sh = shell('juju list-clouds --format yaml')
        if sh.code > 0:
            raise JujuNotFoundException(
                "Unable to list clouds: {}".format(sh.errors())
            )
        return yaml.safe_load("\n".join(sh.output()))

    @classmethod
    def cloud(cls, name):
        """ Return specific cloud information

        Arguments:
        name: name of cloud to query, ie. aws, lxd, local:provider
        Returns:
        Dictionary of cloud attributes
        """
        if name in cls.clouds().keys():
            return cls.clouds()[name]
        raise JujuCloudNotFound("Unable to locate cloud: {}".format(name))

    @classmethod
    def switch(cls, model):
        """ Switch to a Juju Model

        Arguments:
        model: Model to select

        Returns:
        False if failed to switch to Juju Model.
        """
        ret = 0 == shell('juju switch {}'.format(model)).code
        if ret:
            cls.login(True)
        return ret

    @classmethod
    def deploy_bundle(cls, bundle):
        """ Juju deploy bundle

        Arguments:
        bundle: Name of bundle to deploy, can be a path to local bundle file or
                charmstore path.
        """
        return shell('juju deploy {}'.format(bundle))

    @classmethod
    def current_controller(cls):
        """ Grabs the current default controller
        """
        env = os.path.join(juju_path(), 'current-controller')
        if not os.path.isfile(env):
            return None
        with open(env) as fp:
            return fp.read().strip()

    @classmethod
    def controller(cls, controller):
        """ Return specific controller

        Arguments:
        controller: controller id
        """
        controllers = cls.controllers()
        if controllers and controller in controllers:
            return controllers[controller]
        return None

    @classmethod
    def controller_info(cls, name=None):
        """ Returns information on current controller

        Arguments:
        name: if set shows info controller, otherwise displays current.
        """
        cmd = 'juju show-controller --format yaml'
        if name is not None:
            cmd += ' {}'.format(name)
        sh = shell(cmd)
        if sh.code > 0:
            raise JujuNotFoundException(
                "Unable to determine controller: {}".format(sh.errors()))
        out = yaml.safe_load("\n".join(sh.output()))
        try:
            return next(iter(out.values()))
        except:
            return out

    @classmethod
    def controllers(cls):
        """ List available controllers

        Returns:
        List of known controllers
        """
        sh = shell('juju list-controllers --format json')
        if sh.code > 0:
            raise JujuNotFoundException(
                "Unable to list controllers: {}".format(sh.errors()))
        env = json.loads(sh.output()[0])
        return env['controllers']

    @classmethod
    def account(cls, controller):
        """ List account information for controller

        Arguments:
        controller: controller id

        Returns:
        Dictionary containing list of accounts for controller and the
        current account in use.

        eg: {'users': [accounts], 'current': 'admin@local'}
        """
        account = {
            'users': None,
            'current': None
        }
        accounts = cls.accounts()
        if accounts and controller in accounts:
            account['users'] = accounts[controller].get('accounts', [])
            account['current'] = accounts[controller].get(
                'current-account', None)
        return account

    @classmethod
    def accounts(cls):
        """ List available accounts

        Returns:
        List of known accounts
        """
        env = os.path.join(juju_path(), 'accounts.yaml')
        if not os.path.isfile(env):
            raise JujuNotFoundException(
                "Unable to find: {}".format(env))
        with open(env, 'r') as c:
            env = yaml.load(c)
            return env['controllers']
        raise JujuControllerNotFound("Unable to find accounts")

    @classmethod
    def model_by_owner(cls, user):
        """ List model associated with user

        Arguments:
        user: username to query

        Returns:
        Dictionary containing model information for user
        """
        models = cls.models()
        for m in models:
            if m['owner'] == user:
                return m
        raise JujuModelUserNotFound(
            "Unable to find user: {}".format(
                user
            ))

    @classmethod
    def model(cls, name):
        """ List information for model

        Arguments:
        name: model name

        Returns:
        Dictionary of model information
        """
        models = cls.models()['models']
        for m in models:
            if m['name'] == name:
                return m
        raise JujuControllerNotFound(
            "Unable to find model: {}".format(name))

    @classmethod
    def models(cls):
        """ List available models

        Returns:
        List of known models
        """
        sh = shell('juju list-models --format yaml')
        if sh.code > 0:
            raise JujuNotFoundException(
                "Unable to list models: {}".format(sh.errors()))
        out = yaml.safe_load("\n".join(sh.output()))
        return out

    @classmethod
    def current_model(cls):
        return cls.models()['current-model']

    @classmethod
    def version(cls):
        """ Returns version of Juju
        """
        sh = shell('juju version')
        if sh.code > 0:
            raise JujuNotFoundException(
                "Unable to get Juju Version".format(sh.errors()))
        out = sh.output()
        if isinstance(out, list):
            return out.pop()
        else:
            return out



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
import logging


from urwid import (AttrMap, Divider, connect_signal, Edit, Padding,
                   Pile, Text, WidgetWrap)

from ubuntui.ev import EventLoop
from ubuntui.widgets.buttons import MenuSelectButton

from bundleplacer.charmstore_api import CharmStoreAPI

log = logging.getLogger('bundleplacer')


class CharmStoreSearchWidget(WidgetWrap):

    def __init__(self, add_cb, charmstore_column, config, series):
        self.add_cb = add_cb
        self.charmstore_column = charmstore_column
        self.config = config
        self.api = CharmStoreAPI(series=series)
        self.search_delay_alarm = None
        self.search_text = ""
        self._search_future = None
        self._search_result = None
        self._popular_results = None
        w = self.build_widgets()
        super().__init__(w)
        # populate popular_results with an initial search
        self.really_search()

    def build_widgets(self):
        self.editbox = Edit(caption=('text', "Search Charm Store: "))

        connect_signal(self.editbox, 'change',
                       self.handle_edit_changed)

        return Padding(AttrMap(self.editbox,
                               'filter', 'filter_focus'), left=2, right=2)

    def _handle_search_done(self, future):
        assert(future == self._search_future)
        self._search_result = self._search_future.result()
        self._search_future = None

        if self._popular_results is None:
            self._popular_results = self._search_result

        # result being None indicates an error, which was handled by
        # handle_search_error.
        if self._search_result:
            br, cr = self._search_result
            self.set_column(br, cr)
        self.charmstore_column.loading = False
        self.charmstore_column.update()

    def set_column(self, bundle_results, charm_results):
        self.charmstore_column.clear_search_results()
        self.charmstore_column.add_results(bundle_results, charm_results)
        self.charmstore_column.update()

    def handle_edit_changed(self, sender, userdata):
        self.search_text = userdata
        self.enqueue_search()
        self.charmstore_column.handle_search_change(userdata)

    def enqueue_search(self):
        if self.search_delay_alarm:
            EventLoop.remove_alarm(self.search_delay_alarm)
        self.search_delay_alarm = EventLoop.set_alarm_in(0.5,
                                                         self.really_search)

    def really_search(self, *args, **kwargs):
        self.search_delay_alarm = None
        self._search_future = self.api.get_matches(self.search_text,
                                                   self.handle_search_error)
        self._search_future.add_done_callback(self._handle_search_done)

    def handle_search_error(self, e):
        self.charmstore_column.handle_error(e)

    def do_add_charm(self, sender):
        self.add_cb(self.search_text, self._search_result)


class CharmWidget(WidgetWrap):
    def __init__(self, charm_dict, add_cb, recommended=False):
        self.charm_dict = charm_dict
        self.add_cb = add_cb
        self.recommended = recommended
        self.header = None
        self.charm_source = charm_dict['Id']
        self.md = charm_dict['Meta']['charm-metadata']
        self.charm_name = self.md['Name']
        w = self.build_widgets()
        super().__init__(w)

    def selectable(self):
        return True

    def set_header(self, header):
        self.header = header
        self._w = self.build_widgets()

    def build_widgets(self):
        self.charm_name = self.md['Name']
        summary = self.md['Summary']
        top = []
        pad = 0
        if self.header:
            top = [Text("\n{}\n".format(self.header))]
        if self.recommended:
            pad = 2
        s = "{} ({})\n  {}\n".format(self.charm_name,
                                     self.charm_source,
                                     summary)
        b = MenuSelectButton(s, on_press=self.handle_press)
        return Pile(top + [AttrMap(Padding(b, left=pad),
                                   'text',
                                   'button_secondary focus')])

    def handle_press(self, button):
        self.add_cb(self.charm_name, self.charm_dict)


class BundleWidget(WidgetWrap):
    def __init__(self, bundle_dict, add_cb):
        self.add_cb = add_cb
        self.bundle_source = bundle_dict['Id']
        self.bundle_dict = bundle_dict
        self.header = None
        self.md = bundle_dict['Meta']['bundle-metadata']
        w = self.build_widgets()
        super().__init__(w)

    def selectable(self):
        return True

    def set_header(self, header):
        self.header = header
        self._w = self.build_widgets()

    def build_widgets(self):
        name_with_rev = self.bundle_source.split("/")[-1]
        bundle_name = "-".join(name_with_rev.split("-")[:-1])

        if 'Machines' in self.md:
            summary = ("{} services "
                       "on {} machines").format(len(self.md['Services']),
                                                len(self.md['Machines']))
        else:
            summary = "{} services".format(len(self.md['Services']))
        s = "{} ({})\n    {}\n".format(bundle_name, self.bundle_source,
                                       summary)

        top = []
        if self.header:
            top = [Text("\n{}\n".format(self.header))]

        b = MenuSelectButton(s, on_press=self.handle_press)
        return Pile(top + [AttrMap(b,
                                   'text',
                                   'button_secondary focus')])

    def handle_press(self, button):
        self.add_cb(self.md)


class CharmstoreColumnUIState(Enum):
    RELATED = 0
    SEARCH_RESULTS = 1


class CharmstoreColumn(WidgetWrap):
    def __init__(self, display_controller, placement_controller,
                 placement_view, metadata_controller):
        self.placement_controller = placement_controller
        self.display_controller = display_controller
        self.placement_view = placement_view
        self.metadata_controller = metadata_controller
        self.state = CharmstoreColumnUIState.RELATED
        self.prev_state = None
        self.current_search_string = ""
        w = self.build_widgets()
        super().__init__(w)
        self._related_charms = []
        self._bundle_results = []
        self._charm_results = []
        self._recommended_widgets = []
        self.loading = True
        self.update()

    def build_widgets(self):
        self.title = Text('')
        self.pile = Pile([self.title])
        return self.pile

    def clear_search_results(self):
        self._bundle_results = []
        self._charm_results = []

    def handle_search_change(self, s):
        if s == "":
            self.state = CharmstoreColumnUIState.RELATED
            self.clear_search_results()
        else:
            self.state = CharmstoreColumnUIState.SEARCH_RESULTS
        self.current_search_string = s
        self.loading = True
        self.update()

    def update(self):
        opts = self.pile.options()
        extra_widgets = []
        if self.metadata_controller.loaded():
            if len(self._recommended_widgets) == 0:
                rec_dicts = [self.metadata_controller.get_charm_info(n)
                             for n in
                             self.metadata_controller.recommended_charm_names]
                self._recommended_widgets = [(CharmWidget(d, self.do_add_charm,
                                                          recommended=True),
                                              opts) for d in rec_dicts]
                if len(self._recommended_widgets) > 0:
                    top_w = self._recommended_widgets[0][0]
                    top_w.set_header("Recommended Charms")
            else:
                self.loading = False

        bundle_widgets = [(BundleWidget(d, self.do_add_bundle),
                           opts) for d in self._bundle_results
                          if 'bundle-metadata' in d.get('Meta', {})]
        if len(bundle_widgets) > 0:
            top_w = bundle_widgets[0][0]
            top_w.set_header("Bundles")

        charm_widgets = [(CharmWidget(d, self.do_add_charm),
                          opts) for d in self._charm_results
                         if 'charm-metadata' in d.get('Meta', {})]
        if len(charm_widgets) > 0:
            top_w = charm_widgets[0][0]
            top_w.set_header("Charms")

        if self.state == CharmstoreColumnUIState.RELATED:
            if self.loading:
                self.title.set_text("\nLoading Recommended "
                                    "and Popular Charms…")
            else:
                self.title.set_text("")
            extra_widgets = self._recommended_widgets
        else:
            if self.loading:
                msg = "\nSearching for '{}'…\n".format(
                    self.current_search_string)
                self.title.set_text(msg)
            else:
                bn = len(self._bundle_results)
                cn = len(self._charm_results)
                if bn + cn == 0:
                    advice = ""
                    if len(self.current_search_string) < 3:
                        advice = "Try a longer search string."
                    msg = ("\nNo charms found matching '{}' "
                           "{}".format(self.current_search_string,
                                         advice))
                else:
                    msg = ("\nShowing the top {} bundles and {} "
                           "charms matching {}:"
                           "".format(bn, cn, self.current_search_string))
            self.title.set_text(msg)

        self.pile.contents[1:] = extra_widgets + bundle_widgets + charm_widgets

    def add_results(self, bundle_results, charm_results):
        self._bundle_results += bundle_results
        self._bundle_results = self._bundle_results[:5]
        existing_charms = [s.charm_source.rsplit('-', 1)[0] for s in
                           self.placement_controller.bundle.services]
        filtered_charm_results = [c for c in charm_results
                                  if c['Id'].rsplit('-', 1)[0]
                                  not in existing_charms]
        self._charm_results += filtered_charm_results
        self._charm_results = self._charm_results[:10]

    def focus_prev_or_top(self):
        if len(self.pile.contents) < 2:
            return
        if self.state == CharmstoreColumnUIState.RELATED and self.loading \
           and len(self.pile.contents) > 2:
            self.pile.focus_position = 2
        else:
            self.pile.focus_position = 1

    def do_add_charm(self, charm_name, charm_dict):
        self.placement_view.do_add_charm(charm_name,
                                         charm_dict)

    def do_add_bundle(self, bundle_dict):
        self.placement_view.do_add_bundle(bundle_dict)

    def handle_error(self, e):
        msg = "Error searching for {}: {}".format(self.current_search_string,
                                                  e)
        self.title.set_text(msg)



from conjure.ui.views.newcloud import NewCloudView
from conjure.models.provider import Schema
from conjure.utils import juju_path, pollinate
from configobj import ConfigObj
import os.path as path
import yaml
import petname


class NewCloudController:
    """ Renders an input view for defining selected clouds credentials
    """

    def __init__(self, app):
        self.app = app

    def _format_creds(self, creds):
        """ Formats the credentials into strings from the widgets values
        """
        formatted = {}
        for k, v in creds.items():
            if k.startswith('_'):
                # Not a widget but a private key
                k = k[1:]
                formatted[k] = v
            elif k.startswith('@'):
                # A Widget, but not stored in credentials
                continue
            else:
                formatted[k] = v.value
        return formatted

    def finish(self, credentials=None, back=False):
        """ Load the Model controller passing along the selected cloud.

        Arguments:
        credentials: credentials to store for provider
        back: if true loads previous controller
        """
        self.app.current_controller = petname.Name()

        if back:
            return self.app.controllers['welcome'].render()

        cred_path = path.join(juju_path(), 'credentials.yaml')
        if not path.isfile(cred_path):
            existing_creds = {
                    'credentials': {
                        self.cloud: {
                            self.app.current_controller: self._format_creds(
                                credentials)
                        }
                    }
                }

        else:
            existing_creds = yaml.safe_load(open(cred_path))

            if self.cloud in existing_creds['credentials'].keys():
                c = existing_creds['credentials'][self.cloud]
                c[self.app.current_controller] = self._format_creds(
                    credentials)
            else:
                # Handle the case where path exists but an entry for the cloud
                # has yet to be added.
                existing_creds['credentials'][self.cloud] = {
                    self.app.current_controller: self._format_creds(
                        credentials)
                }

        with open(cred_path, 'w') as cred_f:
            cred_f.write(yaml.safe_dump(existing_creds,
                                        default_flow_style=False))

        if self.cloud == 'maas':
            self.cloud = '{}/{}'.format(self.cloud,
                                        credentials['@maas-server'].value)
        pollinate(self.app.session_id, 'CA', self.app.log)
        self.app.controllers['jujucontroller'].render(
            self.cloud, bootstrap=True)

    def render(self, cloud):
        """ Render

        Arguments:
        cloud: The cloud to create credentials for
        """

        # LXD is a special case as we want to make sure a bridge
        # is configured. If not we'll bring up a new view to allow
        # a user to configure a LXD bridge with suggested network
        # information.
        if cloud == 'lxd':
            if path.isfile('/etc/default/lxd-bridge'):
                cfg = ConfigObj('/etc/default/lxd-bridge')
            else:
                cfg = ConfigObj()

            ready = cfg.get('LXD_IPV4_ADDR', None)
            if not ready:
                return self.app.controllers['lxdsetup'].render()

            self.app.log.debug("Found an IPv4 address ({}), "
                               "assuming LXD is configured.".format(ready))
            return self.app.controllers['jujucontroller'].render(
                cloud='lxd', bootstrap=True)

        try:
            creds = Schema[cloud]
        except KeyError as e:
            pollinate(self.app.session_id, 'EC', self.app.log)
            self.app.ui.show_exception_message(e)

        self.config = self.app.config
        self.cloud = cloud
        self.view = NewCloudView(self.app,
                                 self.cloud,
                                 creds,
                                 self.finish)

        self.app.ui.set_header(
            title="New cloud setup",
        )
        self.app.ui.set_body(self.view)



from urwid import WidgetWrap, Text, Filler, Pile
from ubuntui.utils import Padding, Color
from ubuntui.widgets.buttons import (cancel_btn)
from ubuntui.ev import EventLoop


class DeployView(WidgetWrap):
    def __init__(self, app, cb):
        self.app = app
        self.text = Text("deploying...")
        _pile = [
            Padding.center_79(self.text),
            Padding.line_break(""),
            Padding.center_20(self.buttons())
        ]
        super().__init__(Filler(Pile(_pile), valign="middle"))

    def buttons(self):
        cancel = cancel_btn(on_press=self.cancel)

        buttons = [
            Color.button_secondary(cancel, focus_map='button_secondary focus')
        ]
        return Pile(buttons)

    def set_status(self, msg):
        self.text.set_text(msg)

    def cancel(self, btn):
        EventLoop.exit(0)



import random
from urwid import (Text, WidgetWrap)
from ubuntui.widgets.juju.service import ServiceWidget
from ubuntui.widgets.table import Table
from ubuntui.utils import Color
from conjure.api.models import model_status


class ServicesView(WidgetWrap):

    view_columns = [
        ('Icon', "", 2),
        ('Name', "Service", 0),
        ('AgentStatus', "Status", 20),
        ('PublicAddress', "IP", 20),
        ('Machine', "Machine", 20),
    ]

    def __init__(self, app):
        self.app = app
        self.deployed = {}
        self.unit_w = None
        self.table = Table()

        headings = []
        for key, label, width in self.view_columns:
            # If no width assume ('weight', 1, widget)
            if width == 0:
                headings.append(Color.column_header(Text(label)))
            else:
                headings.append(
                    ('fixed', width, Color.column_header(Text(label))))
        self.table.addHeadings(headings)
        super().__init__(self.table.render())

        self.refresh_nodes()

    def refresh_nodes(self):
        """ Adds services to the view if they don't already exist
        """
        status = model_status()
        for name, service in status['Services'].items():
            service_w = ServiceWidget(name, service)
            for unit in service_w.Units:
                services_list = []
                try:
                    unit_w = self.deployed[unit._name]
                except:
                    self.deployed[unit._name] = unit
                    unit_w = self.deployed[unit._name]
                    for k, label, width in self.view_columns:
                        if width == 0:
                            services_list.append(getattr(unit_w, k))
                        else:
                            if not hasattr(unit_w, k):
                                continue
                            services_list.append(('fixed', width,
                                                  getattr(unit_w, k)))

                    self.table.addColumns(unit._name, services_list)
                    if not hasattr(unit_w, 'WorkloadInfo'):
                        continue
                    self.table.addColumns(
                        unit._name,
                        [
                            ('fixed', 5, Text("")),
                            Color.info_context(
                                unit_w.WorkloadInfo)
                        ],
                        force=True)
                self.update_ui_state(unit_w, unit._unit)

    def status_icon_state(self, agent_state):
        if agent_state == "maintenance" \
           or agent_state == "allocating" \
           or agent_state == "executing":
            pending_status = [("pending_icon", "\N{CIRCLED BULLET}"),
                              ("pending_icon", "\N{CIRCLED WHITE BULLET}"),
                              ("pending_icon", "\N{FISHEYE}")]
            status = random.choice(pending_status)
        elif agent_state == "waiting":
            status = ("pending_icon", "\N{HOURGLASS}")
        elif agent_state == "idle" \
                or agent_state == "active":
            status = ("success_icon", "\u2713")
        elif agent_state == "blocked":
            status = ("error_icon", "\N{BLACK FLAG}")
        elif agent_state == "unknown":
            status = ("error_icon", "\N{DOWNWARDS BLACK ARROW}")
        else:
            # NOTE: Should not get here, if we do make sure we account
            # for that error type above.
            status = ("error_icon", "?")
        return status

    def update_ui_state(self, unit_w, unit):
        """ Updates individual machine information

        Arguments:
        service: current service
        unit_w: UnitInfo widget
        unit: current unit for service
        """
        try:
            unit_w.Machine.set_text(unit.get('Machine', '-'))
            unit_w.PublicAddress.set_text(unit['PublicAddress'])
            unit_w.WorkloadInfo.set_text(unit['WorkloadStatus']['Info'])
            if unit['WorkloadStatus']['Status'] != 'unknown':
                unit_w.AgentStatus.set_text(unit['WorkloadStatus']['Status'])
                unit_w.Icon.set_text(
                    self.status_icon_state(unit['WorkloadStatus']['Status']))
            else:
                unit_w.AgentStatus.set_text(unit['AgentStatus']['Status'])
                unit_w.Icon.set_text(
                    self.status_icon_state(unit['AgentStatus']['Status']))
        except Exception as e:
            self.app.log.exception(e)
            self.app.ui.show_exception_message(e)



#
# Copyright 2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum, IntEnum, unique
import logging

log = logging.getLogger('bundleplacer')


class InstallState(IntEnum):
    RUNNING = 0
    NODE_WAIT = 1


@unique
class ControllerState(IntEnum):

    """Names for current screen state"""
    INSTALL_WAIT = 0
    PLACEMENT = 1
    SERVICES = 2
    ADD_SERVICES = 3


class ServiceState(Enum):

    """ Service interaction states """
    REQUIRED = 0
    OPTIONAL = 1
    CONFLICTED = 2



from conjure.api.models import model_info
from conjure.charm import get_bundle
from conjure.models.bundle import BundleModel
from conjure.utils import pollinate
from conjure.juju import Juju, current_controller

from bundleplacer.config import Config
from bundleplacer.maas import connect_to_maas
from bundleplacer.placerview import PlacerView
from bundleplacer.controller import PlacementController, BundleWriter


class DeployController:

    def __init__(self, app):
        self.app = app
        self.placement_controller = None
        self.bundle = None

    def finish(self, back=False):
        """ handles deployment

        Arguments:
        back: if true returns to previous controller
        """
        if back:
            return self.app.controllers['jujucontroller'].render()

        bw = BundleWriter(self.placement_controller)
        bw.write_bundle(self.bundle)
        pollinate(self.app.session_id, 'PC', self.app.log)
        self.app.controllers['deploysummary'].render(self.bundle)

    def render(self, model):
        self.app.current_model = model
        info = model_info(self.app.current_model)

        # Set our provider type environment var so that it is
        # exposed in future processing tasks
        self.app.env['JUJU_PROVIDERTYPE'] = info['ProviderType']

        # Grab bundle and deploy or render placement if MAAS
        self.bundle = get_bundle(BundleModel.to_entity(), to_file=True)
        metadata_filename = self.app.config['metadata_filename']

        bundleplacer_cfg = Config(
            'bundle-placer',
            {'bundle_filename': self.bundle,
             'metadata_filename': metadata_filename,
             'config_filename': self.app.argv.build_conf,
             'bundle_key': BundleModel.key(),
             'provider_type': info['ProviderType']})

        if info['ProviderType'] == 'maas':
            pollinate(self.app.session_id, 'PM', self.app.log)
            try:
                controller_meta = Juju.controller_info(current_controller())
                bootstrap_config = controller_meta['bootstrap-config']
                self.app.log.debug(
                    'bootstrap_config {}'.format(bootstrap_config))
            except Exception as e:
                msg = ("Unable to query cache file, trying "
                       "alternate api: {}".format(e))
                self.app.log.error(msg)
                return self.app.ui.show_exception_message(Exception(msg))

            maasoauth = Juju.credential(bootstrap_config['cloud-type'],
                                        bootstrap_config['credential'])
            # add maas creds to env
            self.app.env['MAAS_SERVER'] = bootstrap_config['region']
            self.app.env['MAAS_OAUTH'] = maasoauth['maas-oauth']

            creds = dict(
                api_host=self.app.env['MAAS_SERVER'],
                api_key=self.app.env['MAAS_OAUTH'])
            maas, maas_state = connect_to_maas(creds)
            self.placement_controller = PlacementController(
                config=bundleplacer_cfg,
                maas_state=maas_state)
            mainview = PlacerView(self.placement_controller,
                                  bundleplacer_cfg,
                                  self.finish, has_maas=True)
            self.app.ui.set_header(
                title=self.app.config['summary'],
                excerpt=("Place services, add additional charms, and manage "
                         "service relations")
            )
            self.app.ui.set_subheader("Bundle Editor")
            self.app.ui.set_body(mainview)
            mainview.update()
        else:
            pollinate(self.app.session_id, 'PS', self.app.log)
            self.placement_controller = PlacementController(
                config=bundleplacer_cfg)
            mainview = PlacerView(self.placement_controller,
                                  bundleplacer_cfg,
                                  self.finish)
            self.app.ui.set_header(
                title=self.app.config['summary'],
                excerpt=("Add additional charms and manage service relations")
            )
            self.app.ui.set_subheader("Bundle Editor")
            self.app.ui.set_body(mainview)
            mainview.update()



""" Project Scaffolding entrypoint
"""

import argparse
import sys
import json
import os
from conjure.shell import shell


class ScaffoldException(Exception):
    """ Error in scaffolding
    """
    pass


class Scaffold:
    def __init__(self, opts):
        """ init

        Arguments:
        opts: Options passed in from cli
        """
        if os.path.isdir(opts.directory):
            raise ScaffoldException(
                "{} exists, please specify another.".format(opts.directory)
            )

        shell('mkdir -p {}'.format(opts.directory))

        config_path = "/usr/share/{}/config.json".format(opts.name)
        if not os.path.isfile(config_path):
            raise ScaffoldException(
                "Unable to find configuration file {}".format(config_path)
            )
        with open(config_path) as configfp:
            config = json.load(configfp)

        for bundle in config['bundles']:
            bundle_key_path = os.path.join(opts.directory, bundle['key'])
            print("Creating {} directory".format(bundle_key_path))
            shell('mkdir -p {}'.format(bundle_key_path))

        # render_lxd_sh(os.path.join(opts.directory, 'lxd.sh'))


def parse_options(argv):
    parser = argparse.ArgumentParser(description="Conjure scaffold",
                                     prog="conjure-scaffold")
    parser.add_argument('-n', '--name', dest='name', metavar='package_name',
                        help='Name of package')
    parser.add_argument('directory',
                        help='directory to scaffold the project in')

    return parser.parse_args(argv)


def main():
    opts = parse_options(sys.argv[1:])

    if not opts.name:
        raise ScaffoldException(
            "A package name is required.")

    if not opts.directory:
        raise ScaffoldException(
            "A project directory is required."
        )

    try:
        Scaffold(opts)
        sys.exit(0)
    except ScaffoldException as e:
        print(e)
        sys.exit(1)



""" Interfaces to Juju API ModelManager """

from conjure.juju import Juju, requires_login


@requires_login
def list_models(user='user-admin'):
    """ Lists Juju Models

    Arguments:
    user: Name of user to list models for.

    Returns:
    Dictionary of known Juju Models (default: user-admin)
    """
    models = Juju.client.ModelManager(request="ListModels",
                                      params={'Tag': user})
    return models['UserModels']


@requires_login
def model_info(model):
    """ Returns information on select model

    Arguments:
    model: name of model to inspect

    Returns:
    Dictionary of model attributes
    """
    return Juju.client.Client(request="ModelInfo",
                              params={"Name": model})


@requires_login
def model_status():
    """ Returns the FullStatus output of a model

    Returns:
    Dictionary of model status
    """
    return Juju.client.Client(request="FullStatus")



from conjure.ui.views.welcome import WelcomeView
from conjure.models.bundle import BundleModel
from conjure.juju import Juju
from conjure.utils import pollinate


class WelcomeController:
    def __init__(self, app):
        self.app = app
        self.view = WelcomeView(self.app, self.finish)

    def finish(self, name):
        """ Finalizes welcome controller

        Arguments:
        name: name of charm/bundle to use
        """
        deploy_key = next((n for n in
                           self.app.config['bundles']
                           if n["name"] == name), None)

        if deploy_key is None:
            raise Exception(
                "Unable to determine bundle to deploy: {}".format(name))

        BundleModel.bundle = deploy_key
        pollinate(self.app.session_id, 'B001', self.app.log)
        if Juju.controllers() is None:
            self.app.controllers['clouds'].render()
        else:
            self.app.controllers['jujucontroller'].render()

    def render(self):
        pollinate(self.app.session_id, 'W001', self.app.log)
        config = self.app.config
        self.app.ui.set_header(
            title=config['summary'],
            excerpt=config['excerpt'],
        )
        self.app.ui.set_body(self.view)



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from urwid import Divider, Pile, Text, WidgetWrap

from bundleplacer.maas import satisfies, MaasMachineStatus

from bundleplacer.ui.filter_box import FilterBox
from bundleplacer.ui.simple_machine_widget import SimpleMachineWidget

log = logging.getLogger('bundleplacer')


class MachinesList(WidgetWrap):

    """A list of machines with configurable action buttons for each
    machine.

    action - a function to call when the machine's button is pressed

    constraints - a dict of constraints to filter the machines list.
    only machines matching all the constraints will be shown.

    show_hardware - bool, whether or not to show the hardware details
    for each of the machines

    title_widgets - A Text Widget to be used in place of the default
    title.

    show_assignments - bool, whether or not to show the assignments
    for each of the machines.

    show_only_ready - bool, only show machines with a ready state.

    """

    def __init__(self, controller, display_controller,
                 constraints=None, show_hardware=False,
                 title_widgets=None, show_assignments=True,
                 show_placeholders=True, show_only_ready=False,
                 show_filter_box=False):
        self.controller = controller
        self.display_controller = display_controller
        self.machine_widgets = []
        if constraints is None:
            self.constraints = {}
        else:
            self.constraints = constraints
        self.show_hardware = show_hardware
        self.show_assignments = show_assignments
        self.show_placeholders = show_placeholders
        self.show_only_ready = show_only_ready
        self.show_filter_box = show_filter_box
        self.filter_string = ""
        w = self.build_widgets(title_widgets)
        self.update()
        super().__init__(w)

    def selectable(self):
        # overridden to ensure that we can arrow through the buttons
        # shouldn't be necessary according to documented behavior of
        # Pile & Columns, but discovered via trial & error.
        return True

    def build_widgets(self, title_widgets):
        if title_widgets is None:
            if len(self.constraints) > 0:
                cstr = " matching constraints"
            else:
                cstr = ""

            title_widgets = [Text("Machines" + cstr, align='center')]

        self.filter_edit_box = FilterBox(self.handle_filter_change)

        header_widgets = title_widgets + [Divider()]

        if self.show_filter_box:
            header_widgets.append(self.filter_edit_box)

        self.header_padding = len(header_widgets)
        self.machine_pile = Pile(header_widgets + self.machine_widgets)
        return self.machine_pile

    def handle_filter_change(self, edit_button, userdata):
        self.filter_string = userdata
        self.update()

    def find_machine_widget(self, m):
        return next((mw for mw in self.machine_widgets if
                     mw.machine.instance_id == m.instance_id), None)

    def update(self):
        machines = self.controller.machines(
            include_placeholders=self.show_placeholders)
        if self.show_only_ready:
            machines = [m for m in machines
                        if m.status == MaasMachineStatus.READY]
        for mw in self.machine_widgets:
            machine = next((m for m in machines if
                            mw.machine.instance_id == m.instance_id), None)
            if machine is None:
                self.remove_machine(mw.machine)

        n_satisfying_machines = len(machines)

        def get_placement_filter_label(d):
            s = ""
            for atype, al in d.items():
                s += " ".join(["{} {}".format(cc.service_name,
                                              cc.display_name)
                               for cc in al])
            return s

        for m in machines:
            if not satisfies(m, self.constraints)[0]:
                self.remove_machine(m)
                n_satisfying_machines -= 1
                continue

            assignment_names = ""
            ad = self.controller.assignments_for_machine(m)
            assignment_names = get_placement_filter_label(ad)
            filter_label = "{} {}".format(m.filter_label(),
                                          assignment_names)
            if self.filter_string != "" and \
               self.filter_string not in filter_label:
                self.remove_machine(m)
                continue

            mw = self.find_machine_widget(m)
            if mw is None:
                mw = self.add_machine_widget(m)
            mw.update()

        self.filter_edit_box.set_info(len(self.machine_widgets),
                                      n_satisfying_machines)

        self.sort_machine_widgets()

    def add_machine_widget(self, machine):
        mw = SimpleMachineWidget(machine,
                                 self.controller,
                                 self.display_controller,
                                 self.show_assignments)
        self.machine_widgets.append(mw)
        options = self.machine_pile.options()
        self.machine_pile.contents.append((mw, options))

        # NOTE: see the +1: indexing in remove_machine if you re-add
        # this divider. it should then be +2.

        # self.machine_pile.contents.append((AttrMap(Padding(Divider('\u23bc'),
        #                                                    left=2, right=2),
        #                                            'label'), options))
        return mw

    def remove_machine(self, machine):
        mw = self.find_machine_widget(machine)
        if mw is None:
            return

        self.machine_widgets.remove(mw)
        mw_idx = 0
        for w, opts in self.machine_pile.contents:
            if w == mw:
                break
            mw_idx += 1

        c = self.machine_pile.contents[:mw_idx] + \
            self.machine_pile.contents[mw_idx + 1:]
        self.machine_pile.contents = c

    def sort_machine_widgets(self):
        def keyfunc(mw):
            m = mw.machine
            hwinfo = " ".join(map(str, [m.arch, m.cpu_cores, m.mem,
                                        m.storage]))
            if str(mw.machine.status) == 'ready':
                skey = 'A'
            else:
                skey = str(mw.machine.status)
            return skey + mw.machine.hostname + hwinfo
        self.machine_widgets.sort(key=keyfunc)

        def wrappedkeyfunc(t):
            mw, options = t
            if not isinstance(mw, SimpleMachineWidget):
                return 'A'
            return keyfunc(mw)

        self.machine_pile.contents.sort(key=wrappedkeyfunc)

    def focus_prev_or_top(self):
        self.update()
        try:
            if self.machine_pile.focus_position <= self.header_padding:
                self.machine_pile.focus_position = self.header_padding
        except IndexError:
            log.debug("index error in machines_list focus_top")



from .api import Base

from functools import partial

# https://github.com/juju/juju/blob/master/api/facadeversions.go
_FACADE_VERSIONS = {
    "Action":                       1,
    "Addresser":                    2,
    "Agent":                        2,
    "AgentTools":                   1,
    "AllWatcher":                   1,
    "AllModelWatcher":              2,
    "Annotations":                  2,
    "Backups":                      1,
    "Block":                        2,
    "Charms":                       2,
    "CharmRevisionUpdater":         1,
    "Client":                       1,
    "Cleaner":                      2,
    "Controller":                   2,
    "Deployer":                     1,
    "DiskManager":                  2,
    "EntityWatcher":                2,
    "FilesystemAttachmentsWatcher": 2,
    "Firewaller":                   2,
    "HighAvailability":             2,
    "ImageManager":                 2,
    "ImageMetadata":                2,
    "InstancePoller":               2,
    "KeyManager":                   1,
    "KeyUpdater":                   1,
    "LeadershipService":            2,
    "Logger":                       1,
    "MachineManager":               2,
    "Machiner":                     1,
    "MetricsManager":               1,
    "MeterStatus":                  1,
    "MetricsAdder":                 2,
    "ModelManager":                 2,
    "Networker":                    1,
    "NotifyWatcher":                1,
    "Pinger":                       1,
    "Provisioner":                  2,
    "ProxyUpdater":                 1,
    "Reboot":                       2,
    "RelationUnitsWatcher":         1,
    "Resumer":                      2,
    "Service":                      3,
    "Storage":                      2,
    "Spaces":                       2,
    "Subnets":                      2,
    "StatusHistory":                2,
    "StorageProvisioner":           2,
    "StringsWatcher":               1,
    "Upgrader":                     1,
    "UnitAssigner":                 1,
    "Uniter":                       3,
    "UserManager":                  1,
    "VolumeAttachmentsWatcher":     2,
    "Undertaker":                   1,
}


class JujuClient(Base):
    """ Exposes Juju 2.0 facades

    Example:
    jujuc = JujuClient(
        'wss://10.0.3.53:17070/model/e712da7b-6808-49ec-8c90-113b26d1650d/api',
        'f2cbbb1f163f2ed8725e973e5eeaf51a')
    jujuc.Client(request="FullStatus")
    """
    API_VERSION = 2

    def __init__(self, url, password, user='user-admin'):
        for name, version in _FACADE_VERSIONS.items():
            setattr(self, name, partial(self._request,
                                        name_type=name,
                                        version=version))
        super().__init__(url, password, user)

    def _request(self, name_type, version, request, params=None):
        """ Performs a request

            Params:
            name_type: Facade type
            version: Facade version
            request: Name of Juju API call
            params: Query options to pass to request
        """
        if params is None:
            params = {}

        if not isinstance(params, dict):
            raise Exception("Must be a dictionary of query parameters.")

        return self.call({'Type': name_type,
                          'Version': version,
                          'Request': request,
                          'Params': params})

    def call(self, params, timeout=None):
        """ Get json data from juju api daemon.

        Params:
        params: Additional params to be passed into request
        """
        with self.connlock:
            req_id = self.conn.do_send(params)

        return self.receive(req_id, timeout)



from conjure.ui.views.lxdsetup import LXDSetupView
from conjure.utils import pollinate, spew
from conjure.shell import shell
from tempfile import NamedTemporaryFile


class LXDSetupController:
    """ Renders a LXD bridge configuration view
    """

    def __init__(self, app):
        self.app = app

    def _format_input(self, network):
        """ Formats the network dictionary into strings from the widgets values
        """
        formatted = {}
        for k, v in network.items():
            widget, help_text = v
            if k.startswith('_'):
                # Not a widget but a private key
                k = k[1:]
            if isinstance(widget.value, bool) and widget.value:
                formatted[k] = str("true")
            elif isinstance(widget.value, bool) and not widget.value:
                formatted[k] = str("false")
            elif widget.value is None:
                formatted[k] = str("")
            else:
                formatted[k] = widget.value
        return formatted

    def _format_conf(self, network):
        """ Formats the lxd bridge config for writing to file
        """
        lines = []
        for k in network.keys():
            lines.append("{}={}".format(k, network[k]))
        return "\n".join(lines)

    def finish(self, lxdnetwork=None, back=False):
        """ Processes the new LXD setup and loads the controller to
        finish bootstrapping the model.

        Arguments:
        back: if true loads previous controller
        """
        if back:
            return self.app.controllers['clouds'].render()

        if lxdnetwork is None:
            return self.app.ui.show_exception_message(
                Exception("Unable to configure LXD network bridge."))

        formatted_network = self._format_input(lxdnetwork)
        self.app.log.debug("LXD Config {}".format(formatted_network))

        out = self._format_conf(formatted_network)

        with NamedTemporaryFile(mode="w", encoding="utf-8",
                                delete=False) as tempf:
            self.app.log.debug("Saving LXD config to {}".format(tempf.name))
            spew(tempf.name, out)
            sh = shell('sudo mv {} /etc/default/lxd-bridge'.format(
                tempf.name))
            if sh.code > 0:
                return self.app.ui.show_exception_message(
                    Exception("Problem saving config: {}".format(sh.errors())))

        self.app.log.debug("Restarting lxd-bridge")
        shell("sudo systemctl restart lxd-bridge.service")

        pollinate(self.app.session_id, 'L002', self.app.log)
        self.app.controllers['jujucontroller'].render(
            cloud='lxd', bootstrap=True)

    def render(self):
        """ Render
        """
        pollinate(self.app.session_id, 'L001', self.app.log)
        self.view = LXDSetupView(self.app,
                                 self.finish)

        self.app.ui.set_header(
            title="Setup LXD Bridge",
        )
        self.app.ui.set_body(self.view)



from __future__ import unicode_literals
from urwid import (WidgetWrap, Text, Pile,
                   Filler, Columns)
from ubuntui.utils import Padding
from ubuntui.widgets.text import Instruction
import random


class BootstrapWaitView(WidgetWrap):

    load_attributes = [('pending_icon', "\u2581"),
                       ('pending_icon', "\u2582"),
                       ('pending_icon', "\u2583"),
                       ('pending_icon', "\u2584"),
                       ('pending_icon', "\u2585"),
                       ('pending_icon', "\u2586"),
                       ('pending_icon', "\u2587"),
                       ('pending_icon', "\u2588")]

    def __init__(self, app):
        self.app = app
        self.message = Instruction('Bootstrapping...', align="center")
        self.loading_boxes = [Text(x) for x in self.load_attributes]
        super().__init__(self._build_node_waiting())

    def redraw_kitt(self):
        """ Redraws the KITT bar
        """
        random.shuffle(self.load_attributes)
        for i in self.loading_boxes:
            i.set_text(
                self.load_attributes[random.randrange(
                    len(self.load_attributes))])

    def _build_node_waiting(self):
        """ creates a loading screen if nodes do not exist yet """
        text = [Padding.line_break(""),
                self.message,
                Padding.line_break("")]

        _boxes = []
        _boxes.append(('weight', 1, Text('')))
        for i in self.loading_boxes:
            _boxes.append(('pack', i))
        _boxes.append(('weight', 1, Text('')))
        _boxes = Columns(_boxes)

        return Filler(Pile(text + [_boxes]),
                      valign="middle")



from conjure.ui.views.jujucontroller import JujuControllerView
from conjure.utils import pollinate
from conjure.juju import Juju
from ubuntui.ev import EventLoop
from conjure.models.bundle import BundleModel
from conjure import async
from functools import partial
import os.path as path
import os
from subprocess import check_output
import json
import petname


class JujuControllerController:
    def __init__(self, app):
        """ init

        Arguments:
        app: common dictionary for conjure
        """
        self.app = app
        self.cloud = None
        self.bootstrap = None
        self._post_bootstrap_pollinate = False

    def handle_exception(self, exc):
        pollinate(self.app.session_id, 'EB', self.app.log)
        self.app.ui.show_exception_message(exc)

    def finish(self, controller=None, back=False):
        """ Deploy to juju controller

        Arguments:
        controller: Juju controller to deploy to
        back: if true returns to previous controller
        """
        self.app.log.debug(
            "jujucontroller - selected: {}".format(controller))
        self.app.current_controller = controller

        if back:
            return self.app.controllers['welcome'].render()

        if self.bootstrap:
            self.app.log.debug("Performing bootstrap: {} {}".format(
                controller, self.cloud))
            future = Juju.bootstrap_async(
                controller=self.app.current_controller,
                cloud=self.cloud,
                series=BundleModel.bootstrapSeries(),
                exc_cb=self.handle_exception,
                log=self.app.log)
            future.add_done_callback(
                self._handle_bootstrap_done)

            self.app.controllers['bootstrapwait'].render()
            pollinate(self.app.session_id, 'J003', self.app.log)

        else:
            self.app.controllers['deploy'].render(self.app.current_controller)

    def _handle_bootstrap_done(self, future):
        self.app.log.debug("handle bootstrap")
        result = future.result()
        if result.code > 0:
            self.app.log.error(result.errors())
            return self.handle_exception(Exception(result.errors()))
        pollinate(self.app.session_id, 'J004', self.app.log)
        EventLoop.remove_alarms()
        Juju.switch(self.app.current_controller)
        self._post_bootstrap_exec()

    def _post_bootstrap_exec(self):
        """ Executes post-bootstrap.sh if exists
        """
        self._post_bootstrap_sh = path.join('/usr/share/',
                                            self.app.config['name'],
                                            'bundles',
                                            BundleModel.key(),
                                            'post-bootstrap.sh')
        if not path.isfile(self._post_bootstrap_sh) \
           or not os.access(self._post_bootstrap_sh, os.X_OK):
            self.app.log.debug(
                "Unable to execute: {}, skipping".format(
                    self._post_bootstrap_sh))
            return self.app.controllers['deploy'].render(
                self.app.current_controller)

        self.app.ui.set_footer('Running post-bootstrap tasks.')

        pollinate(self.app.session_id, 'J001', self.app.log)

        self.app.log.debug("post_bootstrap running: {}".format(
            self._post_bootstrap_sh
        ))

        try:
            future = async.submit(partial(check_output,
                                          self._post_bootstrap_sh,
                                          shell=True,
                                          env=self.app.env),
                                  self.handle_exception)
            future.add_done_callback(self._post_bootstrap_done)
        except Exception as e:
            return self.handle_exception(e)

    def _post_bootstrap_done(self, future):
        try:
            result = json.loads(future.result().decode('utf8'))
        except Exception as e:
            self.handle_exception(e)

        self.app.log.debug("post_bootstrap_done: {}".format(result))
        if result['returnCode'] > 0:
            pollinate(self.app.session_id, 'E001', self.app.log)
            return self.handle_exception(Exception(
                'There was an error during the post '
                'bootstrap processing phase: {}.'.format(result)))
        pollinate(self.app.session_id, 'J002', self.app.log)
        Juju.switch(self.app.current_controller)
        self.app.controllers['deploy'].render(self.app.current_controller)

    def render(self, cloud=None, bootstrap=None):
        """ Render controller

        Arguments:
        cloud: defined cloud to use when deploying
        bootstrap: is this a new environment that needs to be bootstrapped
        """

        self.cloud = cloud

        # Set provider type for post-bootstrap
        self.app.env['JUJU_PROVIDERTYPE'] = self.cloud

        self.bootstrap = bootstrap

        if self.cloud and self.bootstrap:
            if self.app.current_controller is not None:
                return self.finish(self.app.current_controller)
            else:
                self.app.current_controller = petname.Name()
                return self.finish(self.app.current_controller)
            return self.handle_exception(Exception(
                "Unable to determine a controller to bootstrap"))
        else:
            controllers = Juju.controllers().keys()
            models = {}
            for c in controllers:
                Juju.switch(c)
                models[c] = Juju.models()
            self.view = JujuControllerView(self.app,
                                           models,
                                           self.finish)

        self.app.ui.set_header(
            title="Juju Model",
        )
        self.app.ui.set_body(self.view)



import setuptools

# All these imported to be added to our distribution
import conjure
import maasclient  # noqa
import macumba  # noqa
import ubuntui  # noqa
import bundleplacer  # noqa


setuptools.setup(
    name="conjure",
    version=conjure.__version__,
    author="Adam Stokes",
    author_email="adam.stokes@ubuntu.com",
    description="Conjure for apt cloud packages",
    url="https://github.com/ubuntu-solutions-engineering/conjure",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "conjure-setup = conjure.app:main"
        ]
    },
    install_requires=open('requirements.txt', 'r').readlines()
)



# -*- mode: python; -*-
#
# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import os
import yaml


log = logging.getLogger('bundleplacer')


class ConfigException(Exception):
    pass


class Config:
    def __init__(self, name, cfg_obj=None, save_backups=True):
        self.name = name
        if cfg_obj is None:
            self._config = {}
        else:
            self._config = cfg_obj
        self.save_backups = save_backups

    def save(self):
        """ Saves configuration """
        try:
            if not os.path.exists(self.cfg_path):
                os.makedirs(self.cfg_path)
            if self.save_backups and os.path.exists(self.cfg_file):
                datestr = datetime.datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
                backup_path = os.path.join(self.cfg_path, "config-backups")
                backupfilename = "{}/config-{}.yaml".format(backup_path,
                                                            datestr)
                os.makedirs(backup_path, exist_ok=True)
                os.rename(self.cfg_file, backupfilename)
            with open(self.cfg_file, 'w') as f:
                f.write(yaml.safe_dump(dict(self._config),
                                       default_flow_style=False))
        except IOError as e:
            raise ConfigException("Unable to save configuration: {}".format(e))

    @property
    def cfg_path(self):
        """ top level configuration path """
        fn = '~/.config/{}'.format(self.name)
        return os.path.expanduser(fn)

    @property
    def cfg_file(self):
        return os.path.join(self.cfg_path, 'config.yaml')

    @classmethod
    def share_path(cls):
        """ Application share path
        """
        return "/usr/share/bundle-placer/share"

    def setopt(self, key, val):
        """ sets config option """
        try:
            self._config[key] = val
            self.save()
        except Exception as e:
            log.error("Failed to set {} in config: {}".format(key, e))

    def getopt(self, key):
        if key in self._config:
            return self._config[key]
        else:
            if hasattr(self, key):
                attr = getattr(self, key)
                return attr() if callable(attr) else attr
            return False



from collections import OrderedDict
from ubuntui.widgets.input import (StringEditor, YesNo, PasswordEditor)

""" Defining the schema

The schema contains attributes for rendering proper credentials for the
different clouds. There are a few cases where additional attributes are
required and some that are required aren't necessarily meant to be edited.

The schema uses a simple single first value for describing the type of
attribute it is. If there is no indicating first character it is considered
public, editable, and viewable by the end user.

The schema contains the following:

- If a key starts with '_' it is private, not editable, but stored in
  the credentials as is.
- If a key starts with '@' it is public, editable, but _not_ stored in
  the credentials file.
- If the key does not start with any sigil it is public, editable, and
  stored in the credentials file.

Examples:

('maas', OrderedDict([
  ('@maas-server', StringEditor()) # Required input not stored.
  ('_auth-type', 'oauth1) # Required, not editable and is stored.
  ('maas-oauth', StringEditor()) # required, editable, and stored.
])
"""

Schema = OrderedDict([
    ('aws', OrderedDict([
        ('_auth-type', 'access-key'),
        ('access-key', StringEditor()),
        ('secret-key', StringEditor())
    ])),
    ('aws-china', OrderedDict([
        ('_auth-type', 'access-key'),
        ('access-key', StringEditor()),
        ('secret-key', StringEditor())
    ])),
    ('aws-gov', OrderedDict([
        ('_auth-type', 'access-key'),
        ('access-key', StringEditor()),
        ('secret-key', StringEditor())
    ])),
    ('maas', OrderedDict([
        ('_auth-type', 'oauth1'),
        ('@maas-server', StringEditor()),
        ('maas-oauth', StringEditor())
    ])),
    ('azure', OrderedDict([
        ('_auth-type', 'userpass'),
        ('application-id', StringEditor()),
        ('subscription-id', StringEditor()),
        ('tenant-id', StringEditor()),
        ('application-password', PasswordEditor()),
        ('location', StringEditor()),
        ('endpoint', StringEditor()),
        ('storage-endpoint', StringEditor()),
        ('storage-account-type', StringEditor()),
        ('storage-account', StringEditor()),
        ('storage-account-key', StringEditor()),
        ('controller-resource-group', StringEditor())
    ])),
    ('azure-china', OrderedDict([
        ('_auth-type', 'userpass'),
        ('application-id', StringEditor()),
        ('subscription-id', StringEditor()),
        ('tenant-id', StringEditor()),
        ('application-password', PasswordEditor()),
        ('location', StringEditor()),
        ('endpoint', StringEditor()),
        ('storage-endpoint', StringEditor()),
        ('storage-account-type', StringEditor()),
        ('storage-account', StringEditor()),
        ('storage-account-key', StringEditor()),
        ('controller-resource-group', StringEditor())
    ])),
    ('google', OrderedDict([
        ('private-key', StringEditor()),
        ('client-id', StringEditor()),
        ('client-email', StringEditor()),
        ('region', StringEditor()),
        ('project-id', StringEditor()),
        ('image-endpoint', StringEditor())
    ])),
    ('cloudsigma', OrderedDict([
        ('username', StringEditor()),
        ('password', PasswordEditor()),
        ('region', StringEditor()),
        ('endpoint', StringEditor())
    ])),
    ('joyent', OrderedDict([
        ('sdc-user', StringEditor()),
        ('sdc-key-id', StringEditor()),
        ('sdc-url', StringEditor(
            default='https://us-west-1.api.joyentcloud.com')),
        ('private-key-path', StringEditor()),
        ('algorithm', StringEditor(default='rsa-sha256'))
    ])),
    ('openstack', OrderedDict([
        ('_auth-type', 'userpass'),
        ('username', StringEditor()),
        ('password', PasswordEditor()),
        ('tenant-name', StringEditor()),
        ('auth-url', StringEditor()),
        ('auth-mode', StringEditor()),
        ('access-key', StringEditor()),
        ('secret-key', StringEditor()),
        ('region', StringEditor()),
        ('use-floating-ip', YesNo()),
        ('use-default-secgroup', YesNo()),
        ('network', StringEditor())
    ])),
    ('rackspace', OrderedDict([
        ('_auth-type', 'userpass'),
        ('username', StringEditor()),
        ('password', PasswordEditor()),
        ('tenant-name', StringEditor()),
        ('auth-url', StringEditor()),
        ('auth-mode', StringEditor()),
        ('access-key', StringEditor()),
        ('secret-key', StringEditor()),
        ('region', StringEditor()),
        ('use-floating-ip', YesNo()),
        ('use-default-secgroup', YesNo()),
        ('network', StringEditor())
    ])),
    ('vsphere', OrderedDict([
        ('datacenter', StringEditor()),
        ('host', StringEditor()),
        ('user', StringEditor()),
        ('password', PasswordEditor()),
        ('external-network', StringEditor())
    ])),
    ('manual', OrderedDict([
        ('bootstrap-host', StringEditor()),
        ('bootstrap-user', StringEditor()),
        ('use-sshstorage', YesNo())
    ])),

])



from conjure.ui.views.deploy_summary import DeploySummaryView
from conjure.utils import pollinate


class DeploySummaryController:
    def __init__(self, app):
        self.app = app

    def finish(self, back=False):
        """ Does the actual deployment and loads the summary controller

        Arguments:
        back: If true will go back to previous controller
        """
        if back:
            return self.app.controllers['deploy'].render(
                self.app.current_model
            )
        else:
            self.app.controllers['finish'].render(self.bundle)

    def render(self, bundle):
        self.bundle = bundle
        self.excerpt = ("Please review the deployment summary before "
                        "proceeding.")
        self.view = DeploySummaryView(self.app,
                                      self.bundle,
                                      self.finish)

        self.app.ui.set_header(
            title="Deploy Summary",
            excerpt=self.excerpt
        )
        self.app.ui.set_body(self.view)
        pollinate(self.app.session_id, 'SS', self.app.log)



#!/usr/bin/env python3
#
# Sync's upstream git repositories into project directory.

import json
import os
import subprocess
import argparse
import sys
import tempfile
import shutil
import glob
from contextlib import contextmanager


@contextmanager
def chdir(directory):
    """Change the current working directory to a different directory for a code
    block and return the previous directory after the block exits. Useful to
    run commands from a specificed directory.

    :param str directory: The directory path to change to for this context.
    """
    cur = os.getcwd()
    try:
        yield os.chdir(directory)
    finally:
        os.chdir(cur)


def parse_options(argv):
    parser = argparse.ArgumentParser(description="sync-repo",
                                     prog="sync-repo")
    parser.add_argument('-m', '--manifest', dest='manifest_path',
                        metavar='manifest',
                        help='Path to repo manifest')
    parser.add_argument('-f', '--force', dest='force', action='store_true',
                        help='Force syncing, this removes any existing '
                        'directories affected.')
    return parser.parse_args(argv)


def read_manifest(src):
    if not os.path.isfile(src):
        raise Exception("Unable to locate {}".format(src))
    with open(src) as fp:
        return json.load(fp)


def globdirs(src):
    for d in glob.iglob(src, recursive=True):
        if os.path.isdir(d):
            print("Found directory: {}".format(d))
            yield d


def copy_dirs(dirs, src, dst, force=False):
    with chdir(src):
        for d in glob.iglob(dirs):
            dst = os.path.join(dst, d)
            print("Syncing {} to {}".format(d, dst))
            if os.path.isdir(dst) and not force:
                raise Exception(
                    "{} exists, use --force to overwrite.".format(
                        dst))
            if os.path.isdir(dst):
                shutil.rmtree(dst)
            shutil.copytree(d, dst)


def git_clone(src, dst):
    print("Cloning: {} to {}".format(src, dst))
    subprocess.check_call(
        'git clone -q {} {}'.format(src, dst), shell=True)


def git_checkout(src, version):
    with chdir(src):
        subprocess.check_call(
            'git checkout -q {}'.format(version), shell=True)


def main():
    opts = parse_options(sys.argv[1:])
    manifest = read_manifest(opts.manifest_path)
    for repo in manifest['repos']:
        with tempfile.TemporaryDirectory() as repo_tmp:
            git_clone(repo['url'], repo_tmp)
            git_checkout(repo_tmp, repo['version'])
            copy_dirs(repo['directories'],
                      repo_tmp,
                      os.path.abspath(os.curdir),
                      opts.force)


if __name__ == "__main__":
    main()



# Copyright 2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from urwid import (AttrMap, Columns, connect_signal, Edit, Pile, Text,
                   WidgetWrap)

log = logging.getLogger('bundleplacer')


class FilterBox(WidgetWrap):
    def __init__(self, edit_changed_cb):
        self.label = Text("")
        self.info_text = Text("")
        self.editbox = Edit(caption=('text', "Filter: "))
        connect_signal(self.editbox, 'change',
                       edit_changed_cb)

        w = Pile([Columns([AttrMap(self.editbox,
                                   'filter', 'filter_focus')])
                  # self.info_text  # -- WORKAROUND for issue #194
                  ])
        super().__init__(w)

    def set_info(self, n_showing, n_total):
        m = ["Filter ", ('label', "({} of {} shown): ".format(n_showing,
                                                              n_total))]
        self.editbox.set_caption(m)
        if False:   # WORKAROUND for issue #194
            t = ''
        else:
            t = ('label',
                 "  Filter on hostname or hardware info like 'cores:4'")
        self.info_text.set_text(t)



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import (Popen, PIPE, call,
                        check_call, DEVNULL, CalledProcessError)
from contextlib import contextmanager
try:
    from collections import Mapping
except ImportError:
    Mapping = dict

from jinja2 import Environment, FileSystemLoader
import os
import re
import string
import random
import fnmatch
import logging
import urwid
import itertools
import configparser
import time
import sys
import errno
import shutil
import json
import yaml
import requests
from urllib.parse import urlparse

log = logging.getLogger('bundleplacer')

# String with number of minutes, or None.
blank_len = None


class UtilsException(Exception):
    pass


def cleanup(cfg):
    # Save latest config object
    log.info("Cleanup, saving latest config object.")
    cfg.save()
    pid = os.path.join(install_home(), '.cloud-install/openstack.pid')
    if os.path.isfile(pid):
        os.remove(pid)
    if not cfg.getopt('headless') and cfg.getopt('gui_started'):
        log.debug('Attempting to reset the terminal')
        sys.stderr.write("\x1b[2J\x1b[H")
        call(['stty', 'sane'])
    return


def write_status_file(status='', msg=''):
    """ Writes out a status file

    :param str status: success or fail
    :param str msg: any error/success output
    """
    status_file = os.path.join(install_home(), '.cloud-install/finished.json')
    spew(status_file, json.dumps(dict(status=status, msg=msg)))


def sanitize_cli_opts(opts):
    """ removes false and null items from argument list """
    return {k: v for k, v in vars(opts).items() if v}


def populate_config(opts):
    """ populate configuration suitable for loading in the config
    object merging in cli options.

    :param opts: argparse Namespace class of options
    """
    cfg_from_cli = sanitize_cli_opts(opts)
    if 'config_file' not in opts:
        return cfg_from_cli

    # Override config items from local config
    if opts.config_file:
        cfg_override = yaml.load(slurp(opts.config_file))
        return merge_dicts(cfg_from_cli, cfg_override)


def merge_dicts(*dicts):
    """
    Return a new dictionary that is the result of merging the arguments
    together.
    In case of conflicts, later arguments take precedence over earlier
    arguments.

    ref:  http://stackoverflow.com/a/8795331/3170835
    """
    updated = {}
    # grab all keys
    keys = set()
    for d in dicts:
        keys = keys.union(set(d))

    for key in keys:
        values = [d[key] for d in dicts if key in d]
        # which ones are mapping types? (aka dict)
        maps = [value for value in values if isinstance(value, Mapping)]
        if maps:
            # if we have any mapping types, call recursively to merge them
            updated[key] = merge_dicts(*maps)
        else:
            # otherwise, just grab the last value we have, since later
            # arguments take precedence over earlier arguments
            updated[key] = values[-1]
    return updated


def render_charm_config(config):
    """ Render a config for setting charm config options

    If a custom charm config is passed on the cli it will
    attempt to merge those additional settings without losing
    any pre-existing charm options.
    """
    charm_conf = load_template('charmconf.yaml')
    template_args = dict(
        install_type=config.getopt('install_type'),
        openstack_password=config.getopt('openstack_password'))

    os_branch = config.getopt('openstack_git_branch')
    os_release = config.getopt('openstack_release')
    if os_branch:
        if os_branch == 'master':
            template_args['openstack_git_branch'] = os_branch
        else:
            template_args['openstack_git_branch'] = "{}/{}".format(os_branch,
                                                                   os_release)
    template_args['openstack_release'] = os_release

    ubuntu_series = config.getopt('ubuntu_series')
    # Per the charm docs, the openstack charms support
    # openstack-origin: cloud:series-release for a subset of possible
    # combinations - in particular, precise-icehouse and
    # trusty-{juno,kilo,liberty}. If you want to use another series,
    # you don't get to pick the openstack release, you have to use the
    # default, e.g. wily-liberty.
    if ubuntu_series == "trusty":
        openstack_release = config.getopt('openstack_release')
        openstack_origin = ("cloud:{}-{}".format(ubuntu_series,
                                                 openstack_release))
        template_args['openstack_origin'] = openstack_origin

    if config.is_single():
        template_args['worker_multiplier'] = '1'

    # add http proxy settings - should not be necessary as juju sets
    # these in the charm execution environment, but required for
    # openstack-origin-git. See: https://launchpad.net/bugs/1472357

    for pk in ['http_proxy', 'https_proxy']:
        pv = config.getopt(pk)
        if pv:
            template_args[pk] = pv

    charm_conf_modified = charm_conf.render(**template_args)
    dest_yaml_path = os.path.join(config.cfg_path, 'charmconf.yaml')
    spew(dest_yaml_path, charm_conf_modified)

    # Check for custom charm options
    charm_conf_custom_file = config.getopt('charm_config_file')
    if charm_conf_custom_file and os.path.exists(charm_conf_custom_file):
        log.debug("Found custom charm config, updating charm settings.")
        charm_conf = yaml.load(slurp(dest_yaml_path))
        charm_conf_custom = yaml.load(
            slurp(config.getopt('charm_config_file')))
        charm_conf_merged = merge_dicts(charm_conf,
                                        charm_conf_custom)
        spew(dest_yaml_path, yaml.safe_dump(
            charm_conf_merged, default_flow_style=False))


def chown(path, user, group=None, recursive=False):
    """
    Change user/group ownership of file

    :param path: path of file or directory
    :param str user: new owner username
    :param str group: new owner group name
    :param bool recursive: set files/dirs recursively

    """
    try:
        if not recursive or os.path.isfile(path):
            shutil.chown(path, user, group)
        else:
            for root, dirs, files in os.walk(path):
                shutil.chown(root, user, group)
                for item in dirs:
                    shutil.chown(os.path.join(root, item), user, group)
                for item in files:
                    shutil.chown(os.path.join(root, item), user, group)
    except OSError as e:
        raise UtilsException(e)


def ensure_locale():
    """
    Makes sure LC_ALL is defined to something sensible
    """
    locale_conf = slurp('/etc/default/locale')
    for line in locale_conf.split('\n'):
        if line.startswith('#'):
            continue
        if "LC_ALL" in line:
            return
    new_locale = "LC_ALL=\"{}\"".format(os.getenv('LANG', 'C'))
    with open('/etc/default/locale', 'a+') as f:
        f.write(new_locale)
    return


def apt_install(pkgs):
    """ runs apt-get install against space separated list of `pkgs`
    """
    ensure_locale()
    cmd = ("DEBIAN_FRONTEND=noninteractive /usr/bin/apt-get -qyf "
           "-o Dpkg::Options::=--force-confdef "
           "-o Dpkg::Options::=--force-confold "
           "install {0}".format(pkgs))
    try:
        check_call(cmd, stdout=DEVNULL, stderr=DEVNULL, shell=True)
    except CalledProcessError as e:
        log.error("Problem with package install: {0}".format(e))
        pass


def get_command_output(command, timeout=None, user_sudo=False):
    """ Execute command through system shell

    :param command: command to run
    :param timeout: (optional) use 'timeout' to limit time. default 300
    :param user_sudo: (optional) sudo into install users env. default False.
    :type command: str
    :returns: {status: returncode, output: stdout, err: stderr}
    :rtype: dict

    .. code::

        # Get output of juju status
        cmd_dict = utils.get_command_output('juju status')
    """
    cmd_env = os.environ.copy()
    # set consistent locale
    cmd_env['LC_ALL'] = 'C'
    if timeout:
        command = "timeout %ds %s" % (timeout, command)

    if user_sudo:
        command = "sudo -E -H -u {0} {1}".format(install_user(), command)

    try:
        p = Popen(command, shell=True,
                  stdout=PIPE, stderr=PIPE,
                  bufsize=-1, env=cmd_env, close_fds=True)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return dict(ret=127, output="", err="")
        else:
            raise e
    stdout, stderr = p.communicate()
    if p.returncode == 126 or p.returncode == 127:
        stdout = bytes()
    if not stderr:
        stderr = bytes()
    return dict(status=p.returncode,
                output=stdout.decode('utf-8'),
                err=stderr.decode('utf-8'))


def poll_until_true(cmd, predicate, frequency, timeout=600,
                    ignore_exceptions=False):
    """run get_command_output(cmd) every frequency seconds, until
    predicate(output) returns True. Timeout after timeout seconds.

    returns True if call eventually succeeded, or False if timeout was
    reached.

    Exceptions raised during get_command_output are handled as per
    ignore_exceptions. If True, they are just logged. If False, they
    are re-raised.

    """
    start_time = time.time()
    frequency_stub = time.time()
    while True:
        # continue if frequency not met
        if time.time() - frequency_stub <= frequency:
            continue
        try:
            output = get_command_output(cmd)
        except Exception as e:
            if not ignore_exceptions:
                raise e
            else:
                log.debug("**Ignoring** exception: {}".format(e))
        if predicate(output):
            return True
        if time.time() - start_time >= timeout:
            return False


def remote_cp(machine_id, src, dst, juju_home):
    log.debug("Remote copying {src} to {dst} on machine {m}".format(
        src=src,
        dst=dst,
        m=machine_id))
    ret = get_command_output(
        "{juju_home} juju scp {src} {m}:{dst}".format(
            juju_home=juju_home, src=src, dst=dst, m=machine_id))
    log.debug("Remote copy result: {r}".format(r=ret))


def remote_run(machine_id, cmds, juju_home):
    if type(cmds) is list:
        cmds = " && ".join(cmds)
    log.debug("Remote running ({cmds}) on machine {m}".format(
        m=machine_id, cmds=cmds))
    ret = get_command_output(
        "{juju_home} juju run "
        "--machine {m} '{cmds}'".format(juju_home=juju_home,
                                        m=machine_id,
                                        cmds=cmds))
    log.debug("Remote run result: {r}".format(r=ret))
    return ret


def get_host_mem():
    """ Get host memory

    Mostly used as a backup if no data can be pulled from
    the normal means in Machine()
    """
    cmd = get_command_output('head -n1 /proc/meminfo')
    out = cmd['output'].rstrip()
    regex = re.compile('^MemTotal:\s+(\d+)\skB')
    match = re.match(regex, out)
    if match:
        mem = match.group(1)
        mem = int(mem) / 1024 / 1024 + 1
        return int(mem)
    else:
        return 0


def get_host_storage():
    """ Get host storage

    LXC doesn't report storage so we pull from host
    """
    cmd = get_command_output('df -B G --total -l --output=avail'
                             ' -x devtmpfs -x tmpfs | tail -n 1'
                             ' | tr -d "G"')
    if not cmd['status']:
        return cmd['output'].lstrip()
    else:
        return 0


def get_host_cpu_cores():
    """ Get host cpu-cores

    A backup if no data can be pulled from
    Machine()
    """
    cmd = get_command_output('nproc')
    if cmd['output']:
        return cmd['output'].strip()
    else:
        return 'N/A'


def partition(pred, iterable):
    """ Returns tuple of allocated and unallocated systems

    :param pred: status predicate
    :type pred: function
    :param iterable: machine data
    :type iterable: list
    :returns: ([allocated], [unallocated])
    :rtype: tuple

    .. code::

        def is_allocated(d):
            allocated_states = ['started', 'pending', 'down']
            return 'charms' in d or d['agent_state'] in allocated_states
        allocated, unallocated = utils.partition(is_allocated,
                                                 [{state: 'pending'}])
    """
    yes, no = [], []
    for i in iterable:
        (yes if pred(i) else no).append(i)
    return (yes, no)


def reset_blanking():
    global blank_len
    if blank_len is not None:
        call(('setterm', '-blank', blank_len))


@contextmanager
def console_blank():
    global blank_len
    try:
        with open('/sys/module/kernel/parameters/consoleblank') as f:
            blank_len = f.read()
    except (IOError, FileNotFoundError):  # NOQA
        blank_len = None
    else:
        # Cannot use anything that captures stdout, because it is needed
        # by the setterm command to write to the console.
        call(('setterm', '-blank', '0'))
        # Convert the interval from seconds to minutes.
        blank_len = str(int(blank_len) // 60)

    yield

    reset_blanking()


def randomString(size=6, chars=string.ascii_uppercase + string.digits):
    """ Generate a random string

    :param size: number of string characters
    :type size: int
    :param chars: range of characters (optional)
    :type chars: str

    :returns: a random string
    :rtype: str
    """
    return ''.join(random.choice(chars) for x in range(size))


def random_password(size=32):
    """ Generate a password

    :param int size: length of password
    """
    out = get_command_output("pwgen -s {}".format(size))
    return out['output'].strip()


def time_string():
    """ Time helper

    :returns: formatted current time string
    :rtype: str
    """
    return time.strftime('%Y-%m-%d %H:%M')


def find(file_pattern, top_dir, max_depth=None, path_pattern=None):
    """generator function to find files recursively. Usage:

    .. code::

        for filename in find("*.properties", "/var/log/foobar"):
            print filename
    """
    if max_depth:
        base_depth = os.path.dirname(top_dir).count(os.path.sep)
        max_depth += base_depth

    for path, dirlist, filelist in os.walk(top_dir):
        if max_depth and path.count(os.path.sep) >= max_depth:
            del dirlist[:]

        if path_pattern and not fnmatch.fnmatch(path, path_pattern):
            continue

        for name in fnmatch.filter(filelist, file_pattern):
            yield os.path.join(path, name)


def load_template(name, path=None):
    """ load template file

    :param str name: name of template file
    :param str path: alternate location of template location
    """
    if path is None:
        path = '/usr/share/openstack/templates'
    env = Environment(
        loader=FileSystemLoader(path))
    return env.get_template(name)


def install_user():
    """ returns sudo user
    """
    user = os.getenv('SUDO_USER', None)
    if not user:
        user = os.getenv('USER', 'root')
    return user


def install_home():
    """ returns installer user home
    """
    return os.path.expanduser("~" + install_user())


def ssh_readkey():
    """ reads ssh key
    """
    with open(ssh_pubkey(), 'r') as f:
        return f.read()


def ssh_genkey():
    """ Generates sshkey
    """
    if not os.path.exists(ssh_privkey()):
        user_sshkey_path = os.path.join(install_home(),
                                        '.ssh/id_rsa')
        cmd = "ssh-keygen -N '' -f {0}".format(user_sshkey_path)
        out = get_command_output(cmd, user_sudo=True)
        if out['status'] != 0:
            raise Exception(
                "Unable to generate key: {0}".format(out['output']))
        get_command_output('sudo chown -R {0} {1}'.format(
            install_user(),
            os.path.join(install_home(), '.ssh')))
        get_command_output('chmod 0644 {0}.pub'.format(user_sshkey_path),
                           user_sudo=True)
    else:
        log.debug('ssh keys exist for this user, they will be used instead.')


def read_ini(path):
    """ Reads a basic INI like file without sections headers.
    Prepends a default section header for querying.
    """
    ini = open(path)
    config = configparser.ConfigParser()
    config.read_file(itertools.chain(['[DEFAULT]'], ini))
    return config


def ssh_pubkey():
    """ returns path of ssh public key
    """
    return os.path.join(install_home(), '.ssh/id_rsa.pub')


def ssh_privkey():
    """ returns path of private key
    """
    return os.path.join(install_home(), '.ssh/id_rsa')


def spew(path, data, owner=None):
    """ Writes data to path

    :param str path: path of file to write to
    :param str data: contents to write
    :param str owner: optional owner of file
    """
    with open(path, 'w') as f:
        f.write(data)
    if owner:
        try:
            chown(path, owner)
        except:
            raise UtilsException("Unable to set ownership of {}".format(path))


def slurp(path):
    """ Reads data from path

    :param str path: path of file
    """
    try:
        with open(path) as f:
            return f.read().strip()
    except IOError:
        raise IOError


def human_to_mb(s):
    """Translates human-readable strings like '10G' to numeric
    megabytes"""

    if len(s) == 0:
        raise Exception("unexpected empty string")

    md = dict(M=1, G=1024, T=1024 * 1024, P=1024 * 1024 * 1024)
    suffix = s[-1]
    if suffix.isalpha():
        return float(s[:-1]) * md[suffix]
    else:
        return float(s)


def mb_to_human(num):
    """Translates float number of bytes into human readable strings."""
    suffixes = ['M', 'G', 'T', 'P']
    if num == 0:
        return '0 B'

    i = 0
    while num >= 1024 and i < len(suffixes) - 1:
        num /= 1024
        i += 1
    return "{:.2f} {}".format(num, suffixes[i])


def format_constraint(k, v):
    vs = str(v)
    if vs.isdecimal():
        vs = mb_to_human(v)
    return "{}={}".format(k, vs)


def make_screen_hicolor(screen):
    """returns a screen to pass to MainLoop init
    with 256 colors.
    """
    screen.set_terminal_properties(256)
    screen.reset_default_terminal_palette()
    return screen


def get_hicolor_screen(palette):
    screen = urwid.raw_display.Screen()
    screen.register_palette(palette)
    return make_screen_hicolor(screen)


def macgen():
    """ generates mac addresses
    """
    mac = [0x00, 0x16, 0x3e,
           random.randint(0x00, 0x7f),
           random.randint(0x00, 0xff),
           random.randint(0x00, 0xff)]
    return ':'.join(map(lambda x: "%02x" % x, mac))


def download_url(url, output_file):
    """ Downloads contents from a URL
    :param str url: HTTP resource
    :param str output_file: path to store downloaded contents
    """
    res = requests.get(url)
    if res.ok:
        spew(output_file, res.content.decode('utf-8'))
    else:
        raise UtilsException("Exception downloading {}:{}".format(
            url, res.content))


def pollinate(session, tag):
    """ fetches random seed

    :param str session: randomly generated session id
    :param str tag: custom tag
    """
    if not os.path.isfile('/usr/bin/pollinate'):
        return
    if tag not in ['IL', 'IS', 'IM', 'DL', 'DM', 'DS']:
        raise UtilsException("Unknown TAG {}".format(tag))

    session = os.getenv('OSI_TESTRUNNER_ID', session)
    agent_str = 'uoi/{}/{}'.format(session, tag)
    try:
        cmd = ("sudo su - -c 'pollinate -q -r --curl-opts "
               "\"--user-agent {}\"'".format(agent_str))
        log.info("pollinate: {}".format(cmd))
        check_call(cmd, shell=True)
    except CalledProcessError as e:
        log.warning("Generating random seed failed: {}".format(e))


def parse_openstack_creds(creds_file):
    """ Parses openstack-{admin,ubuntu}-rc for openstack
    credentials
    """
    def sanitize_quotes(key):
        if key[0] == key[-1] and key[0] in ("'", '"'):
            return key[1:-1]
        else:
            return key

    cred_map = {}
    rc = slurp(creds_file)
    lines = rc.split("\n")
    for k in lines:
        first, rest = k.split("=")
        first = first.lower()
        if 'username' in first:
            cred_map['username'] = sanitize_quotes(rest)
        if 'password' in first:
            cred_map['password'] = sanitize_quotes(rest)
        if 'tenant_name' in first:
            cred_map['tenant_name'] = sanitize_quotes(rest)
        if 'auth_url' in first:
            cred_map['auth_url'] = urlparse(sanitize_quotes(rest))
        if 'region_name' in first:
            cred_map['region_name'] = sanitize_quotes(rest)
    return cred_map



#
# auth.py - MAAS Authentication
#
# Copyright 2014 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import check_output
import os
import logging

log = logging.getLogger(__name__)


class MaasAuth:

    """ MAAS Authorization class
    """

    def __init__(self, api_url=None, api_key=None):
        """ Initialize with optional OAuth credentials
        """
        self.api_url = api_url if api_url else 'http://localhost/MAAS/api/1.0'
        self.api_key = api_key
        self.consumer_secret = ''

    @property
    def is_logged_in(self):
        """ Checks if we are logged into the MAAS api

        :rtype: bool
        """
        return True if self.api_key else False

    @property
    def consumer_key(self):
        """ Maas consumer key

        :rtype: str
        """
        return self.api_key.split(':')[0] if self.api_key else None

    @property
    def token_key(self):
        """ Maas oauth token key

        :rtype: str
        """
        return self.api_key.split(':')[1] if self.api_key else None

    @property
    def token_secret(self):
        """ Maas oauth token secret

        :rtype: str
        """
        return self.api_key.split(':')[2] if self.api_key else None

    def get_api_key(self, username='root'):
        """ MAAS api key

        :param username: (optional) MAAS user to query for credentials
        :type username: str
        """
        if os.path.isfile('/usr/sbin/maas-region-admin'):
            out = check_output(['sudo', 'maas-region-admin', 'apikey',
                                '--username', username])
            self.api_key = out.decode('ascii').split()[-1].rstrip('\n')
        else:
            raise Exception(
                'Unable to find maas-region-admin to query api key.')



# Copyright 2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import json

from bundleplacer.maas import MaasMachine

log = logging.getLogger('bundleplacer')


class FakeMaasState:
    """ A fake MAAS fixture for quickly testing bundle placement
    against a set of machines
    """

    server_hostname = "fake.maas"

    def machines(self, state=None, constraints=None):
        fakepath = '/usr/share/bundle-placer/share'
        fn = os.path.join(fakepath, "maas-machines.json")
        if not os.path.exists(fn):
            fn = os.path.join("share", "maas-machines.json")
        with open(fn) as f:
            try:
                nodes = json.load(f)
            except:
                log.exception("Error loading JSON")
                return []
        return [MaasMachine(-1, m) for m in nodes
                if m['hostname'] != 'juju-bootstrap.maas']

    def invalidate_nodes_cache(self):
        "no op"

    def machines_summary(self):
        return "no summary for fake state"

    def get_server_config(self, param):
        return dict(maas_name='fake maas')



# Copyright 2014-2016 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


def label_to_atype(labels):
    """ Maps labels in metadata to assignment types

    Arguments:
    labels: List of assignment types ['LXC', 'KVM', 'BareMetal']

    Returns:
    List of mapped enumerated assignment types
    """
    atypes = []
    for label in labels:
        if isinstance(label, AssignmentType):
            atypes.append(label)
        if label.lower() == "lxc":
            atypes.append(AssignmentType.LXC)
        elif label.lower() == "baremetal":
            atypes.append(AssignmentType.BareMetal)
        elif label.lower() == "kvm":
            atypes.append(AssignmentType.KVM)
        elif label.lower() == "lxd":
            atypes.append(AssignmentType.LXD)
        else:
            return [AssignmentType.DEFAULT]
    return atypes


class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class AssignmentType(OrderedEnum):
    # both are equivalent to not specifying a type to juju:
    DEFAULT = 1
    BareMetal = 1
    KVM = 2
    LXC = 3
    LXD = 4
